# pylint: disable=unused-argument
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, FilePath, validator
from pydantic.generics import GenericModel

from starlite.constants import DEFAULT_CHUNK_SIZE
from starlite.datastructures import ETag
from starlite.datastructures.background_tasks import BackgroundTask, BackgroundTasks
from starlite.datastructures.cookie import Cookie
from starlite.enums import MediaType
from starlite.exceptions import ImproperlyConfiguredException
from starlite.response import (
    FileResponse,
    RedirectResponse,
    StreamingResponse,
    TemplateResponse,
)
from starlite.types import FileInfo, FileSystemProtocol
from starlite.types.composite_types import StreamType
from starlite.utils.file import BaseLocalFileSystem

if TYPE_CHECKING:
    from starlite.app import Starlite
    from starlite.connection import Request

R = TypeVar("R")


class ResponseContainer(ABC, GenericModel, Generic[R]):
    """Generic response container."""

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    background: Optional[Union[BackgroundTask, BackgroundTasks]] = None
    """A [BackgroundTask][starlite.datastructures.BackgroundTask] instance or.

    [BackgroundTasks][starlite.datastructures.BackgroundTasks] to execute after the response is finished.
    Defaults to None.
    """
    headers: Dict[str, Any] = {}
    """A string/string dictionary of response headers.

    Header keys are insensitive. Defaults to None.
    """
    cookies: List[Cookie] = []
    """A list of Cookie instances to be set under the response 'Set-Cookie' header.

    Defaults to None.
    """
    media_type: Optional[Union[MediaType, str]] = None
    """If defined, overrides the media type configured in the route decorator."""
    encoding: str = "utf-8"
    """The encoding to be used for the response headers."""

    @abstractmethod
    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> "R":  # pragma: no cover
        """Abstract method that should be implemented by subclasses.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A Response Object
        """
        raise NotImplementedError("not implemented")


class File(ResponseContainer[FileResponse]):
    """Container type for returning File responses."""

    path: FilePath
    """Path to the file to send."""
    filename: Optional[str] = None
    """An optional filename to set in the header."""
    stat_result: Optional[os.stat_result] = None
    """An optional result of calling 'os.stat'.

    If not provided, this will be done by the response constructor.
    """
    chunk_size: int = DEFAULT_CHUNK_SIZE
    """The size of chunks to use when streaming the file."""
    content_disposition_type: Literal["attachment", "inline"] = "attachment"
    """The type of the 'Content-Disposition'.

    Either 'inline' or 'attachment'.
    """
    etag: Optional[ETag] = None
    """An optional [ETag][starlite.datastructures.ETag] instance.

    If not provided, an etag will be automatically generated.
    """
    file_system: Any = BaseLocalFileSystem()
    """The file_system spec to use loading the file.

    Notes:
        - A file_system is a class that adheres to the
            [FileSystemProtocol][starlite.types.FileSystemProtocol].
        - You can use any of the file systems exported from the
            [fsspec](https://filesystem-spec.readthedocs.io/en/latest/) library for this purpose.
    """
    file_info: Optional[FileInfo] = None
    """The output of calling `file_system.info(..)`, equivalent to providing a `stat_result`."""

    @validator("stat_result", always=True)
    def validate_status_code(  # pylint: disable=no-self-argument
        cls, value: Optional[os.stat_result], values: Dict[str, Any]
    ) -> os.stat_result:
        """Set the stat_result value for the given filepath.

        Args:
            value: An optional result [stat][os.stat] result.
            values: The dict of values.

        Returns:
            A stat_result
        """
        return value or Path(cast("str", values.get("path"))).stat()

    @validator("file_system", always=True)
    def validate_file_system(  # pylint: disable=no-self-argument
        cls, value: "FileSystemProtocol"
    ) -> "FileSystemProtocol":
        """Ensure the value is a file system spec.

        Args:
            value: A file system spec.

        Returns:
            A file system spec.
        """
        if not (callable(getattr(value, "info", None)) and callable(getattr(value, "open", None))):
            raise ValueError("file_system must adhere to the FileSystemProtocol type")
        return value

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Optional[Union["MediaType", str]],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> FileResponse:
        """Create a FileResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A FileResponse instance
        """
        return FileResponse(
            background=self.background,
            chunk_size=self.chunk_size,
            content_disposition_type=self.content_disposition_type,
            encoding=self.encoding,
            etag=self.etag,
            file_info=self.file_info,
            file_system=self.file_system,
            filename=self.filename,
            headers=headers,
            media_type=media_type,
            path=self.path,
            stat_result=self.stat_result,
            status_code=status_code,
        )


class Redirect(ResponseContainer[RedirectResponse]):
    """Container type for returning Redirect responses."""

    path: str
    """Redirection path."""

    def to_response(  # type: ignore[override]
        self,
        headers: Dict[str, Any],
        # TODO: update the redirect response to support HTML as well.
        #   This argument is currently ignored.
        media_type: Union["MediaType", str],
        status_code: Literal[301, 302, 303, 307, 308],
        app: "Starlite",
        request: "Request",
    ) -> RedirectResponse:
        """Create a RedirectResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A RedirectResponse instance
        """
        return RedirectResponse(
            background=self.background,
            encoding=self.encoding,
            headers=headers,
            status_code=status_code,
            url=self.path,
        )


class Stream(ResponseContainer[StreamingResponse]):
    """Container type for returning Stream responses."""

    iterator: Union[StreamType[Union[str, bytes]], Callable[[], StreamType[Union[str, bytes]]]]
    """Iterator, Iterable,Generator or async Iterator, Iterable or Generator returning chunks to stream."""

    @validator("iterator", always=True)
    def validate_iterator(  # pylint: disable=no-self-argument
        cls,
        value: Union[StreamType[Union[str, bytes]], Callable[[], StreamType[Union[str, bytes]]]],
    ) -> StreamType[Union[str, bytes]]:
        """Set the iterator value by ensuring that the return value is iterable.

        Args:
            value: An iterable or callable returning an iterable.

        Returns:
            A sync or async iterable.
        """
        return value if isinstance(value, (Iterable, Iterator, AsyncIterable, AsyncIterator)) else value()

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> StreamingResponse:
        """Create a StreamingResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A StreamingResponse instance
        """

        return StreamingResponse(
            background=self.background,
            content=self.iterator if isinstance(self.iterator, (Iterable, AsyncIterable)) else self.iterator(),
            encoding=self.encoding,
            headers=headers,
            media_type=media_type,
            status_code=status_code,
        )


class Template(ResponseContainer[TemplateResponse]):
    """Container type for returning Template responses."""

    name: str
    """Path-like name for the template to be rendered, e.g. "index.html"."""
    context: Dict[str, Any] = {}
    """A dictionary of key/value pairs to be passed to the temple engine's render method.

    Defaults to None.
    """

    def to_response(
        self,
        headers: Dict[str, Any],
        media_type: Union["MediaType", str],
        status_code: int,
        app: "Starlite",
        request: "Request",
    ) -> "TemplateResponse":
        """Create a TemplateResponse instance.

        Args:
            headers: A dictionary of headers.
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum.
            status_code: A response status code.
            app: The [Starlite][starlite.app.Starlite] application instance.
            request: A [Request][starlite.connection.request.Request] instance.

        Raises:
            [ImproperlyConfiguredException][starlite.exceptions.ImproperlyConfiguredException]: if app.template_engine
                is not configured.

        Returns:
            A TemplateResponse instance
        """
        if not app.template_engine:
            raise ImproperlyConfiguredException("Template engine is not configured")

        return TemplateResponse(
            background=self.background,
            context=self.create_template_context(request=request),
            encoding=self.encoding,
            headers=headers,
            status_code=status_code,
            template_engine=app.template_engine,
            template_name=self.name,
            media_type=media_type,
        )

    def create_template_context(self, request: "Request") -> Dict[str, Any]:
        """Create a context object for the template.

        Args:
            request: A [Request][starlite.connection.request.Request] instance.

        Returns:
            A dictionary holding the template context
        """
        csrf_token = request.scope.get("_csrf_token", "")
        return {
            **self.context,
            "request": request,
            "csrf_input": f'<input type="hidden" name="_csrf_token" value="{csrf_token}" />',
        }



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Type,
    Union,
    cast,
)

from pydantic import BaseConfig, BaseModel, root_validator, validator

from starlite.config.logging import BaseLoggingConfig, LoggingConfig
from starlite.datastructures.state import State  # noqa: TC001
from starlite.exceptions import MissingDependencyException
from starlite.types import BeforeMessageSendHookHandler
from starlite.utils import AsyncCallable
from starlite.utils.serialization import decode_json, encode_json

from .types import SessionMakerInstanceProtocol, SessionMakerTypeProtocol

try:
    from sqlalchemy import create_engine
    from sqlalchemy.engine import Engine
    from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, create_async_engine
    from sqlalchemy.future import Engine as FutureEngine
    from sqlalchemy.orm import Query, Session, sessionmaker
    from sqlalchemy.pool import Pool

except ImportError as e:
    raise MissingDependencyException("sqlalchemy is not installed") from e

if TYPE_CHECKING:

    from starlite.types import Message, Scope

IsolationLevel = Literal["AUTOCOMMIT", "READ COMMITTED", "READ UNCOMMITTED", "REPEATABLE READ", "SERIALIZABLE"]

SESSION_SCOPE_KEY = "_sql_alchemy_db_session"
SESSION_TERMINUS_ASGI_EVENTS = {
    "http.response.start",
    "http.disconnect",
    "websocket.disconnect",
    "websocket.close",
}


def serializer(value: Any) -> str:
    """Serialize JSON field values.

    Args:
        value: Any json serializable value.

    Returns:
        JSON string.
    """
    return encode_json(value).decode("utf-8")


async def default_before_send_handler(message: "Message", _: "State", scope: "Scope") -> None:
    """Handle closing and cleaning up sessions before sending.

    Args:
        message: ASGI-`Message`
        _: A `State` (not used)
        scope: An ASGI-`Scope`

    Returns:
        None
    """
    session = cast("Optional[Union[Session, AsyncSession]]", scope.get(SESSION_SCOPE_KEY))
    if session and message["type"] in SESSION_TERMINUS_ASGI_EVENTS:
        if isinstance(session, AsyncSession):
            await session.close()
        else:
            session.close()
        del scope[SESSION_SCOPE_KEY]  # type: ignore


class SQLAlchemySessionConfig(BaseModel):
    """Configuration for a SQLAlchemy-Session."""

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    autocommit: Optional[bool] = None
    autoflush: Optional[bool] = None
    bind: Optional[Any] = None
    binds: Optional[Any] = None
    enable_baked_queries: Optional[bool] = None
    expire_on_commit: bool = False
    future: Optional[bool] = None
    info: Optional[Dict[str, Any]] = None
    query_cls: Optional[Type[Query]] = None
    twophase: Optional[bool] = None


class SQLAlchemyEngineConfig(BaseModel):
    """Configuration for SQLAlchemy's `Engine`.

    For details see: https://docs.sqlalchemy.org/en/14/core/engines.html
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    connect_args: Optional[Dict[str, Any]] = None
    echo: Optional[bool] = None
    echo_pool: Optional[bool] = None
    enable_from_linting: Optional[bool] = None
    future: bool = True
    hide_parameters: Optional[bool] = None
    isolation_level: Optional[IsolationLevel] = None
    json_deserializer: Callable[[str], Any] = decode_json
    json_serializer: Callable[[Any], str] = serializer
    label_length: Optional[int] = None
    listeners: Any = None
    logging_level: Optional[Union[int, str]] = None
    logging_name: Optional[str] = None
    max_identifier_length: Optional[int] = None
    max_overflow: Optional[int] = None
    module: Any = None
    paramstyle: Optional[Literal["qmark", "numeric", "named", "format", "pyformat"]] = None
    plugins: Optional[List[str]] = None
    pool: Optional[Pool] = None
    pool_logging_name: Optional[str] = None
    pool_pre_ping: Optional[bool] = None
    pool_recycle: Optional[int] = None
    pool_reset_on_return: Optional[Literal["rollback", "commit"]] = None
    pool_size: Optional[int] = None
    pool_timeout: Optional[int] = None
    pool_use_lifo: Optional[bool] = None
    poolclass: Optional[Type[Pool]] = None
    query_cache_size: Optional[int] = None
    strategy: Optional[str] = None


class SQLAlchemyConfig(BaseModel):
    """Configuration for SQLAlchemy's `sessionmaker`.

    For details see: https://docs.sqlalchemy.org/en/14/orm/session_api.html
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    connection_string: Optional[str] = None
    """Database connection string in one of the formats supported by SQLAlchemy.

    Notes:
    - For async connections, the connection string must include the correct async prefix.
        e.g. 'postgresql+asyncpg://...' instead of 'postgresql://', and for sync connections its the opposite.
    """
    use_async_engine: bool = True
    """Dictate whether the engine created is an async connection or not.

    Notes:
    - This option must correlate to the type of 'connection_string'. That is, an async connection string required an
        async connection and vice versa.
    """
    create_async_engine_callable: Callable[[str], AsyncEngine] = create_async_engine
    """Callable that creates an 'AsyncEngine' instance or instance of its subclass."""
    create_engine_callable: Callable[[str], Union[Engine, FutureEngine]] = create_engine
    """Callable that creates an 'Engine' or 'FutureEngine' instance or instance of its subclass."""
    dependency_key: str = "db_session"
    """Key to use for the dependency injection of database sessions."""
    engine_app_state_key: str = "db_engine"
    """Key under which to store the SQLAlchemy engine in the application [State][starlite.datastructures.State]
    instance.
    """
    engine_config: SQLAlchemyEngineConfig = SQLAlchemyEngineConfig()
    """Configuration for the SQLAlchemy engine.

    The configuration options are documented in the SQLAlchemy documentation.
    """
    set_json_serializers: bool = True
    """A boolean flag dictating whether to set 'msgspec' based serializer/deserializer functions.

    Notes:
    - Some databases or some versions of some databases do not have a JSON column type. E.g. some older versions of
        SQLite for example. In this case this flag should be false or an error will be raised by SQLAlchemy.
    """
    session_class: Optional[Union[Type[Session], Type[AsyncSession]]] = None
    """The session class to use.

    If not set, the session class will default to 'sqlalchemy.orm.Session' for sync connections and
    'sqlalchemy.ext.asyncio.AsyncSession' for async ones.
    """
    session_config: SQLAlchemySessionConfig = SQLAlchemySessionConfig()
    """Configuration options for the 'sessionmaker'.

    The configuration options are documented in the SQLAlchemy documentation.
    """
    session_maker_class: Type[SessionMakerTypeProtocol] = sessionmaker
    """Sessionmaker class to use."""
    session_maker_app_state_key: str = "session_maker_class"
    """Key under which to store the SQLAlchemy 'sessionmaker' in the application [State][starlite.datastructures.State]
    instance.
    """
    session_maker_instance: Optional[SessionMakerInstanceProtocol] = None
    """Optional sessionmaker to use.

    If set, the plugin will use the provided instance rather than instantiate a sessionmaker.
    """
    engine_instance: Optional[Union[Engine, FutureEngine, AsyncEngine]] = None
    """Optional engine to use.

    If set, the plugin will use the provided instance rather than instantiate an engine.
    """
    before_send_handler: BeforeMessageSendHookHandler = default_before_send_handler
    """Handler to call before the ASGI message is sent.

    The handler should handle closing the session stored in the ASGI scope, if its still open, and committing and
    uncommitted data.
    """

    @validator("before_send_handler", always=True)
    def validate_before_send_handler(  # pylint: disable=no-self-argument
        cls, value: BeforeMessageSendHookHandler
    ) -> Any:
        """Wrap `before_send_handler` in an `AsyncCallable`

        Args:
            value: A before send handler callable.

        Returns:
            An `AsyncCallable`
        """
        return AsyncCallable(value)  # type: ignore[arg-type]

    @root_validator
    def check_connection_string_or_engine_instance(  # pylint: disable=no-self-argument
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Either `connection_string` or `engine_instance` must be specified, and not both.

        Args:
            values: Field values, after validation.

        Returns:
            Field values.
        """
        connection_string = values.get("connection_string")
        engine_instance = values.get("engine_instance")

        if connection_string is None and engine_instance is None:
            raise ValueError("One of 'connection_string' or 'engine_instance' must be provided.")

        if connection_string is not None and engine_instance is not None:
            raise ValueError("Only one of 'connection_string' or 'engine_instance' can be provided.")

        return values

    @property
    def engine_config_dict(self) -> Dict[str, Any]:
        """Return the engine configuration as a dict.

        Returns:
            A string keyed dict of config kwargs for the SQLAlchemy 'create_engine' function.
        """
        engine_excluded_fields: Set[str] = {"future", "logging_level"} if self.use_async_engine else {"logging_level"}

        if not self.set_json_serializers:
            engine_excluded_fields.update({"json_deserializer", "json_serializer"})

        return self.engine_config.dict(exclude_none=True, exclude=engine_excluded_fields)

    @property
    def engine(self) -> Union["Engine", "FutureEngine", "AsyncEngine"]:
        """Return an engine. If none exists yet, create one.

        Returns:
            Getter that returns the engine instance used by the plugin.
        """
        if not self.engine_instance:
            create_engine_callable = (
                self.create_async_engine_callable if self.use_async_engine else self.create_engine_callable
            )
            self.engine_instance = create_engine_callable(
                self.connection_string, **self.engine_config_dict  # type:ignore[arg-type]
            )
        return cast("Union[Engine, FutureEngine, AsyncEngine]", self.engine_instance)

    @property
    def session_maker(self) -> sessionmaker:
        """Get a sessionmaker. If none exists yet, create one.

        Returns:
            Getter that returns the session_maker instance used by the plugin.
        """
        if not self.session_maker_instance:
            session_maker_kwargs = self.session_config.dict(
                exclude_none=True, exclude={"future"} if self.use_async_engine else set()
            )
            session_class = self.session_class or (AsyncSession if self.use_async_engine else Session)
            self.session_maker_instance = self.session_maker_class(
                self.engine, class_=session_class, **session_maker_kwargs
            )
        return cast("sessionmaker", self.session_maker_instance)

    def create_db_session_dependency(self, state: State, scope: "Scope") -> Union[Session, AsyncSession]:
        """Create a session instance.

        Args:
            state: The 'application.state' instance.
            scope: The current connection's scope.

        Returns:
            A session instance T.
        """
        session = scope.get(SESSION_SCOPE_KEY)
        if not session:
            session_maker = cast("sessionmaker", state[self.session_maker_app_state_key])
            session = scope[SESSION_SCOPE_KEY] = session_maker()  # type: ignore
        return cast("Union[Session, AsyncSession]", session)

    def update_app_state(self, state: State) -> None:
        """Create a DB engine and stores it in the application state.

        Args:
            state: The 'application.state' instance.

        Returns:
            None
        """

        state[self.engine_app_state_key] = self.engine
        state[self.session_maker_app_state_key] = self.session_maker

    async def on_shutdown(self, state: State) -> None:
        """Disposes of the SQLAlchemy engine.

        Args:
            state: The 'application.state' instance.

        Returns:
            None
        """
        engine = cast("Union[Engine, AsyncEngine]", state[self.engine_app_state_key])
        if isinstance(engine, AsyncEngine):
            await engine.dispose()
        else:
            engine.dispose()
        del state[self.engine_app_state_key]

    def config_sql_alchemy_logging(self, logging_config: Optional[BaseLoggingConfig]) -> None:
        """Add the SQLAlchemy loggers to the logging config.

        Notes:
            - Currently only works with [LoggingConfig][starlite.config.logging.LoggingConfig].

        Args:
            logging_config: Logging config.

        Returns:
            None.
        """
        if isinstance(logging_config, LoggingConfig):
            logger_settings = {
                "level": self.engine_config.logging_level or "WARNING",
                "handlers": logging_config.loggers["starlite"]["handlers"],
            }
            for logger in (
                "sqlalchemy",
                self.engine_config.logging_name or "sqlalchemy.engine",
                self.engine_config.pool_logging_name or "sqlalchemy.pool",
            ):
                if logger not in logging_config.loggers:
                    logging_config.loggers[logger] = logger_settings



from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Literal,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from typing_extensions import TypedDict

from starlite.connection.request import Request
from starlite.datastructures.upload_file import UploadFile
from starlite.enums import HttpMethod, RequestEncodingType
from starlite.parsers import parse_cookie_string

if TYPE_CHECKING:
    from starlite.connection import ASGIConnection
    from starlite.types import Method
    from starlite.types.asgi_types import HTTPResponseBodyEvent, HTTPResponseStartEvent


def obfuscate(values: Dict[str, Any], fields_to_obfuscate: Set[str]) -> Dict[str, Any]:
    """Obfuscate values in a dictionary, replacing values with `******`

    Args:
        values: A dictionary of strings
        fields_to_obfuscate: keys to obfuscate

    Returns:
        A dictionary with obfuscated strings
    """
    for key in values:
        if key.lower() in fields_to_obfuscate:
            values[key] = "*****"
    return values


RequestExtractorField = Literal[
    "path", "method", "content_type", "headers", "cookies", "query", "path_params", "body", "scheme", "client"
]

ResponseExtractorField = Literal["status_code", "headers", "body", "cookies"]


class ExtractedRequestData(TypedDict, total=False):
    """Dictionary representing extracted request data."""

    body: Coroutine
    client: Tuple[str, int]
    content_type: Tuple[str, Dict[str, str]]
    cookies: Dict[str, str]
    headers: Dict[str, str]
    method: "Method"
    path: str
    path_params: Dict[str, Any]
    query: Union[bytes, Dict[str, Any]]
    scheme: str


class ConnectionDataExtractor:
    """Utility class to extract data from an.

    [ASGIConnection][starlite.connection.ASGIConnection],
    [Request][starlite.connection.Request] or [WebSocket][starlite.connection.WebSocket] instance.
    """

    __slots__ = (
        "connection_extractors",
        "request_extractors",
        "parse_body",
        "parse_query",
        "obfuscate_headers",
        "obfuscate_cookies",
    )

    def __init__(
        self,
        extract_body: bool = True,
        extract_client: bool = True,
        extract_content_type: bool = True,
        extract_cookies: bool = True,
        extract_headers: bool = True,
        extract_method: bool = True,
        extract_path: bool = True,
        extract_path_params: bool = True,
        extract_query: bool = True,
        extract_scheme: bool = True,
        obfuscate_cookies: Optional[Set[str]] = None,
        obfuscate_headers: Optional[Set[str]] = None,
        parse_body: bool = False,
        parse_query: bool = False,
    ):
        """Initialize `ConnectionDataExtractor`

        Args:
            extract_body: Whether to extract body, (for requests only).
            extract_client: Whether to extract the client (host, port) mapping.
            extract_content_type: Whether to extract the content type and any options.
            extract_cookies: Whether to extract cookies.
            extract_headers: Whether to extract headers.
            extract_method: Whether to extract the HTTP method, (for requests only).
            extract_path: Whether to extract the path.
            extract_path_params: Whether to extract path parameters.
            extract_query: Whether to extract query parameters.
            extract_scheme: Whether to extract the http scheme.
            obfuscate_headers: headers keys to obfuscate. Obfuscated values are replaced with '*****'.
            obfuscate_cookies: cookie keys to obfuscate. Obfuscated values are replaced with '*****'.
            parse_body: Whether to parse the body value or return the raw byte string, (for requests only).
            parse_query: Whether to parse query parameters or return the raw byte string.
        """
        self.parse_body = parse_body
        self.parse_query = parse_query
        self.obfuscate_headers = {h.lower() for h in (obfuscate_headers or set())}
        self.obfuscate_cookies = {c.lower() for c in (obfuscate_cookies or set())}
        self.connection_extractors: Dict[str, Callable[["ASGIConnection[Any, Any, Any]"], Any]] = {}
        self.request_extractors: Dict[RequestExtractorField, Callable[["Request[Any, Any]"], Any]] = {}
        if extract_scheme:
            self.connection_extractors["scheme"] = self.extract_scheme
        if extract_client:
            self.connection_extractors["client"] = self.extract_client
        if extract_path:
            self.connection_extractors["path"] = self.extract_path
        if extract_headers:
            self.connection_extractors["headers"] = self.extract_headers
        if extract_cookies:
            self.connection_extractors["cookies"] = self.extract_cookies
        if extract_query:
            self.connection_extractors["query"] = self.extract_query
        if extract_path_params:
            self.connection_extractors["path_params"] = self.extract_path_params
        if extract_method:
            self.request_extractors["method"] = self.extract_method
        if extract_content_type:
            self.request_extractors["content_type"] = self.extract_content_type
        if extract_body:
            self.request_extractors["body"] = self.extract_body

    def __call__(self, connection: "ASGIConnection[Any, Any, Any]") -> ExtractedRequestData:
        """Extract data from the connection, returning a dictionary of values.

        Notes:
            - The value for 'body' - if present - is an unresolved Coroutine and as such should be awaited by the receiver.

        Args:
            connection: An ASGI connection or its subclasses.

        Returns:
            A string keyed dictionary of extracted values.
        """
        extractors = (
            {**self.connection_extractors, **self.request_extractors}  # type: ignore
            if isinstance(connection, Request)
            else self.connection_extractors
        )
        return cast("ExtractedRequestData", {key: extractor(connection) for key, extractor in extractors.items()})

    @staticmethod
    def extract_scheme(connection: "ASGIConnection[Any, Any, Any]") -> str:
        """Extract the scheme from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            The connection's scope["scheme"] value
        """
        return connection.scope["scheme"]

    @staticmethod
    def extract_client(connection: "ASGIConnection[Any, Any, Any]") -> Tuple[str, int]:
        """Extract the client from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            The connection's scope["client"] value or a default value.
        """
        return connection.scope.get("client") or ("", 0)

    @staticmethod
    def extract_path(connection: "ASGIConnection[Any, Any, Any]") -> str:
        """Extract the path from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            The connection's scope["path"] value
        """
        return connection.scope["path"]

    def extract_headers(self, connection: "ASGIConnection[Any, Any, Any]") -> Dict[str, str]:
        """Extract headers from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            A dictionary with the connection's headers.
        """
        headers = {k.decode("latin-1"): v.decode("latin-1") for k, v in connection.scope["headers"]}
        return obfuscate(headers, self.obfuscate_headers) if self.obfuscate_headers else headers

    def extract_cookies(self, connection: "ASGIConnection[Any, Any, Any]") -> Dict[str, str]:
        """Extract cookies from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            A dictionary with the connection's cookies.
        """
        return obfuscate(connection.cookies, self.obfuscate_cookies) if self.obfuscate_cookies else connection.cookies

    def extract_query(self, connection: "ASGIConnection[Any, Any, Any]") -> Any:
        """Extract query from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            Either a dictionary with the connection's parsed query string or the raw query byte-string.
        """
        return connection.query_params.dict() if self.parse_query else connection.scope.get("query_string", b"")

    @staticmethod
    def extract_path_params(connection: "ASGIConnection[Any, Any, Any]") -> Dict[str, Any]:
        """Extract the path parameters from an `ASGIConnection`

        Args:
            connection: An [ASGIConnection][starlite.connection.ASGIConnection] instance.

        Returns:
            A dictionary with the connection's path parameters.
        """
        return connection.path_params

    @staticmethod
    def extract_method(request: "Request[Any, Any]") -> "Method":
        """Extract the method from an `ASGIConnection`

        Args:
            request: A [Request][starlite.connection.Request] instance.

        Returns:
            The request's scope["method"] value.
        """
        return request.scope["method"]

    @staticmethod
    def extract_content_type(request: "Request[Any, Any]") -> Tuple[str, Dict[str, str]]:
        """Extract the content-type from an `ASGIConnection`

        Args:
            request: A [Request][starlite.connection.Request] instance.

        Returns:
            A tuple containing the request's parsed 'Content-Type' header.
        """
        return request.content_type

    async def extract_body(self, request: "Request[Any, Any]") -> Any:
        """Extract the body from an `ASGIConnection`

        Args:
            request: A [Request][starlite.connection.Request] instance.

        Returns:
            Either the parsed request body or the raw byte-string.
        """
        if request.method != HttpMethod.GET:
            if not self.parse_body:
                return await request.body()
            request_encoding_type = request.content_type[0]
            if request_encoding_type == RequestEncodingType.JSON:
                return await request.json()
            form_data = await request.form()
            if request_encoding_type == RequestEncodingType.URL_ENCODED:
                return dict(form_data)
            return {
                key: repr(value) if isinstance(value, UploadFile) else value for key, value in form_data.multi_items()
            }
        return None


class ExtractedResponseData(TypedDict, total=False):
    """Dictionary representing extracted response data."""

    body: bytes
    status_code: int
    headers: Dict[str, str]
    cookies: Dict[str, str]


class ResponseDataExtractor:
    """Utility class to extract data from a `Message`"""

    __slots__ = ("extractors", "parse_headers", "obfuscate_headers", "obfuscate_cookies")

    def __init__(
        self,
        extract_body: bool = True,
        extract_cookies: bool = True,
        extract_headers: bool = True,
        extract_status_code: bool = True,
        obfuscate_cookies: Optional[Set[str]] = None,
        obfuscate_headers: Optional[Set[str]] = None,
    ):
        """Initialize `ResponseDataExtractor` with options.

        Args:
            extract_body: Whether to extract the body.
            extract_cookies: Whether to extract the cookies.
            extract_headers: Whether to extract the headers.
            extract_status_code: Whether to extract the status code.
            obfuscate_cookies: cookie keys to obfuscate. Obfuscated values are replaced with '*****'.
            obfuscate_headers: headers keys to obfuscate. Obfuscated values are replaced with '*****'.
        """
        self.obfuscate_headers = {h.lower() for h in (obfuscate_headers or set())}
        self.obfuscate_cookies = {c.lower() for c in (obfuscate_cookies or set())}
        self.extractors: Dict[
            ResponseExtractorField, Callable[[Tuple["HTTPResponseStartEvent", "HTTPResponseBodyEvent"]], Any]
        ] = {}
        if extract_body:
            self.extractors["body"] = self.extract_response_body
        if extract_status_code:
            self.extractors["status_code"] = self.extract_status_code
        if extract_headers:
            self.extractors["headers"] = self.extract_headers
        if extract_cookies:
            self.extractors["cookies"] = self.extract_cookies

    def __call__(self, messages: Tuple["HTTPResponseStartEvent", "HTTPResponseBodyEvent"]) -> ExtractedResponseData:
        """Extract data from the response, returning a dictionary of values.

        Args:
            messages: A tuple containing
                [HTTPResponseStartEvent][starlite.types.asgi_types.HTTPResponseStartEvent]
                and [HTTPResponseBodyEvent][starlite.types.asgi_types.HTTPResponseBodyEvent].

        Returns:
            A string keyed dictionary of extracted values.
        """
        return cast("ExtractedResponseData", {key: extractor(messages) for key, extractor in self.extractors.items()})

    @staticmethod
    def extract_response_body(messages: Tuple["HTTPResponseStartEvent", "HTTPResponseBodyEvent"]) -> bytes:
        """Extract the response body from a `Message`

        Args:
            messages: A tuple containing
                [HTTPResponseStartEvent][starlite.types.asgi_types.HTTPResponseStartEvent]
                and [HTTPResponseBodyEvent][starlite.types.asgi_types.HTTPResponseBodyEvent].

        Returns:
            The Response's body as a byte-string.
        """
        return messages[1]["body"]

    @staticmethod
    def extract_status_code(messages: Tuple["HTTPResponseStartEvent", "HTTPResponseBodyEvent"]) -> int:
        """Extract a status code from a `Message`

        Args:
            messages: A tuple containing
                [HTTPResponseStartEvent][starlite.types.asgi_types.HTTPResponseStartEvent]
                and [HTTPResponseBodyEvent][starlite.types.asgi_types.HTTPResponseBodyEvent].

        Returns:
            The Response's status-code.
        """
        return messages[0]["status"]

    def extract_headers(self, messages: Tuple["HTTPResponseStartEvent", "HTTPResponseBodyEvent"]) -> Dict[str, str]:
        """Extract headers from a `Message`

        Args:
            messages: A tuple containing
                [HTTPResponseStartEvent][starlite.types.asgi_types.HTTPResponseStartEvent]
                and [HTTPResponseBodyEvent][starlite.types.asgi_types.HTTPResponseBodyEvent].

        Returns:
            The Response's headers dict.
        """
        headers = {
            key.decode("latin-1"): value.decode("latin-1")
            for key, value in filter(lambda x: x[0].lower() != b"set-cookie", messages[0]["headers"])
        }
        return (
            obfuscate(
                headers,
                self.obfuscate_headers,
            )
            if self.obfuscate_headers
            else headers
        )

    def extract_cookies(self, messages: Tuple["HTTPResponseStartEvent", "HTTPResponseBodyEvent"]) -> Dict[str, str]:
        """Extract cookies from a `Message`

        Args:
            messages: A tuple containing
                [HTTPResponseStartEvent][starlite.types.asgi_types.HTTPResponseStartEvent]
                and [HTTPResponseBodyEvent][starlite.types.asgi_types.HTTPResponseBodyEvent].

        Returns:
            The Response's cookies dict.
        """
        cookie_string = ";".join(
            list(  # noqa: C417
                map(
                    lambda x: x[1].decode("latin-1"),
                    filter(lambda x: x[0].lower() == b"set-cookie", messages[0]["headers"]),
                )
            )
        )
        if cookie_string:
            parsed_cookies = parse_cookie_string(cookie_string)
            return obfuscate(parsed_cookies, self.obfuscate_cookies) if self.obfuscate_cookies else parsed_cookies
        return {}



from copy import copy
from http import HTTPStatus
from operator import attrgetter
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Tuple, Type

from pydantic_openapi_schema.v3_1_0 import Response
from pydantic_openapi_schema.v3_1_0.header import Header
from pydantic_openapi_schema.v3_1_0.media_type import (
    MediaType as OpenAPISchemaMediaType,
)
from pydantic_openapi_schema.v3_1_0.schema import Schema
from typing_extensions import get_args, get_origin

from starlite.datastructures.response_containers import File, Redirect, Stream, Template
from starlite.enums import MediaType
from starlite.exceptions import (
    HTTPException,
    ImproperlyConfiguredException,
    ValidationException,
)
from starlite.openapi.enums import OpenAPIFormat, OpenAPIType
from starlite.openapi.schema import create_schema
from starlite.openapi.utils import pascal_case_to_text
from starlite.response import Response as StarliteResponse
from starlite.utils import (
    create_parsed_model_field,
    get_enum_string_value,
    get_name,
    is_class_and_subclass,
)

if TYPE_CHECKING:

    from pydantic_openapi_schema.v3_1_0.responses import Responses

    from starlite.datastructures.cookie import Cookie
    from starlite.handlers import HTTPRouteHandler
    from starlite.plugins.base import PluginProtocol


def create_cookie_schema(cookie: "Cookie") -> Schema:
    """Given a Cookie instance, return its corresponding OpenAPI schema.

    Args:
        cookie: Cookie

    Returns:
        Schema
    """
    cookie_copy = copy(cookie)
    cookie_copy.value = "<string>"
    value = cookie_copy.to_header(header="")
    return Schema(description=cookie.description or "", example=value)


def create_success_response(
    route_handler: "HTTPRouteHandler", generate_examples: bool, plugins: List["PluginProtocol"]
) -> Response:
    """Create the schema for a success response."""
    signature = route_handler.signature
    default_descriptions: Dict[Any, str] = {
        Stream: "Stream Response",
        Redirect: "Redirect Response",
        File: "File Download",
    }
    description = (
        route_handler.response_description
        or default_descriptions.get(signature.return_annotation)
        or HTTPStatus(route_handler.status_code).description
    )
    if signature.return_annotation not in {signature.empty, None, Redirect, File, Stream}:
        return_annotation = signature.return_annotation
        if signature.return_annotation is Template:
            return_annotation = str  # since templates return str
            route_handler.media_type = get_enum_string_value(MediaType.HTML)
        elif is_class_and_subclass(get_origin(signature.return_annotation), StarliteResponse):
            return_annotation = get_args(signature.return_annotation)[0] or Any
        as_parsed_model_field = create_parsed_model_field(return_annotation)
        schema = create_schema(field=as_parsed_model_field, generate_examples=generate_examples, plugins=plugins)
        schema.contentEncoding = route_handler.content_encoding
        schema.contentMediaType = route_handler.content_media_type
        response = Response(
            content={
                route_handler.media_type: OpenAPISchemaMediaType(
                    media_type_schema=schema,
                )
            },
            description=description,
        )
    elif signature.return_annotation is Redirect:
        response = Response(
            content=None,
            description=description,
            headers={
                "location": Header(
                    param_schema=Schema(type=OpenAPIType.STRING), description="target path for the redirect"
                )
            },
        )
    elif signature.return_annotation in (File, Stream):
        response = Response(
            content={
                route_handler.media_type: OpenAPISchemaMediaType(
                    media_type_schema=Schema(
                        type=OpenAPIType.STRING,
                        contentEncoding=route_handler.content_encoding or "application/octet-stream",
                        contentMediaType=route_handler.content_media_type,
                    ),
                )
            },
            description=description,
            headers={
                "content-length": Header(
                    param_schema=Schema(type=OpenAPIType.STRING), description="File size in bytes"
                ),
                "last-modified": Header(
                    param_schema=Schema(type=OpenAPIType.STRING, schema_format=OpenAPIFormat.DATE_TIME),
                    description="Last modified data-time in RFC 2822 format",
                ),
                "etag": Header(param_schema=Schema(type=OpenAPIType.STRING), description="Entity tag"),
            },
        )
    else:
        response = Response(
            content=None,
            description=description,
        )
    if response.headers is None:
        response.headers = {}
    for key, value in route_handler.resolve_response_headers().items():
        header = Header()
        for attribute_name, attribute_value in value.dict(exclude_none=True).items():
            if attribute_name == "value":
                model_field = create_parsed_model_field(type(attribute_value))
                header.param_schema = create_schema(field=model_field, generate_examples=False, plugins=plugins)
            elif attribute_name != "documentation_only":
                setattr(header, attribute_name, attribute_value)
        response.headers[key] = header
    cookies = route_handler.resolve_response_cookies()
    if cookies:
        response.headers["Set-Cookie"] = Header(
            param_schema=Schema(
                allOf=[create_cookie_schema(cookie=cookie) for cookie in sorted(cookies, key=attrgetter("key"))]
            )
        )
    return response


def create_error_responses(exceptions: List[Type[HTTPException]]) -> Iterator[Tuple[str, Response]]:
    """Create the schema for error responses, if any."""
    grouped_exceptions: Dict[int, List[Type[HTTPException]]] = {}
    for exc in exceptions:
        if not grouped_exceptions.get(exc.status_code):
            grouped_exceptions[exc.status_code] = []
        grouped_exceptions[exc.status_code].append(exc)
    for status_code, exception_group in grouped_exceptions.items():
        exceptions_schemas = [
            Schema(
                type=OpenAPIType.OBJECT,
                required=["detail", "status_code"],
                properties={
                    "status_code": Schema(type=OpenAPIType.INTEGER),
                    "detail": Schema(type=OpenAPIType.STRING),
                    "extra": Schema(
                        type=[OpenAPIType.NULL, OpenAPIType.OBJECT, OpenAPIType.ARRAY], additionalProperties=Schema()
                    ),
                },
                description=pascal_case_to_text(get_name(exc)),
                examples=[{"status_code": status_code, "detail": HTTPStatus(status_code).phrase, "extra": {}}],
            )
            for exc in exception_group
        ]
        if len(exceptions_schemas) > 1:
            schema = Schema(oneOf=exceptions_schemas)  # type:ignore[arg-type]
        else:
            schema = exceptions_schemas[0]
        yield str(status_code), Response(
            description=HTTPStatus(status_code).description,
            content={MediaType.JSON: OpenAPISchemaMediaType(media_type_schema=schema)},
        )


def create_additional_responses(
    route_handler: "HTTPRouteHandler", plugins: List["PluginProtocol"]
) -> Iterator[Tuple[str, Response]]:
    """Create the schema for additional responses, if any."""
    if not route_handler.responses:
        return

    for status_code, additional_response in route_handler.responses.items():
        model_field = create_parsed_model_field(additional_response.model)
        schema = create_schema(
            field=model_field, generate_examples=additional_response.generate_examples, plugins=plugins
        )
        yield str(status_code), Response(
            description=additional_response.description,
            content={additional_response.media_type: OpenAPISchemaMediaType(media_type_schema=schema)},
        )


def create_responses(
    route_handler: "HTTPRouteHandler",
    raises_validation_error: bool,
    generate_examples: bool,
    plugins: List["PluginProtocol"],
) -> Optional["Responses"]:
    """Create a Response model embedded in a `Responses` dictionary for the given RouteHandler or return None."""
    responses: "Responses" = {
        str(route_handler.status_code): create_success_response(
            route_handler=route_handler,
            generate_examples=generate_examples,
            plugins=plugins,
        ),
    }
    exceptions = route_handler.raises or []
    if raises_validation_error and ValidationException not in exceptions:
        exceptions.append(ValidationException)
    for status_code, response in create_error_responses(exceptions=exceptions):
        responses[status_code] = response

    for status_code, response in create_additional_responses(route_handler, plugins):
        if status_code in responses:
            raise ImproperlyConfiguredException(
                f"Additional response for status code {status_code} already exists in success or error responses"
            )

        responses[status_code] = response

    return responses or None



from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, TypeVar, Union, cast

T = TypeVar("T")

if TYPE_CHECKING:
    from starlite.types import MaybePartial


def get_name(value: Any) -> str:
    """Get the `__name__` of an object.

    Args:
        value: An arbitrary object.

    Returns:
        A name string.
    """
    if hasattr(value, "__name__"):
        return cast("str", value.__name__)
    return type(value).__name__


def get_enum_string_value(value: Union[Enum, str]) -> str:
    """Return the string value of a string enum.

    See: https://github.com/starlite-api/starlite/pull/633#issuecomment-1286519267

    Args:
        value: An enum or string.

    Returns:
        A string.
    """
    return value.value if isinstance(value, Enum) else value  # type:ignore


@dataclass
class Ref(Generic[T]):
    """A helper class that encapsulates a value."""

    __slots__ = ("value",)

    value: T
    """The value wrapped by the ref."""


def unwrap_partial(value: "MaybePartial[T]") -> T:
    """Unwraps a partial, returning the underlying callable.

    Args:
        value: A partial function.

    Returns:
        Callable
    """
    output: Any = value.func if hasattr(value, "func") else value  # pyright: ignore
    while hasattr(output, "func"):
        output = output.func
    return cast("T", output)



from mimetypes import guess_type
from pathlib import PurePath
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from starlite.enums import MediaType
from starlite.response.base import Response
from starlite.status_codes import HTTP_200_OK

if TYPE_CHECKING:
    from starlite.datastructures import BackgroundTask, BackgroundTasks
    from starlite.template import TemplateEngineProtocol
    from starlite.types import ResponseCookies


class TemplateResponse(Response[bytes]):
    """Template-based response, rendering a given template into a bytes string."""

    def __init__(
        self,
        template_name: str,
        *,
        template_engine: "TemplateEngineProtocol",
        context: Dict[str, Any],
        status_code: int = HTTP_200_OK,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
        encoding: str = "utf-8",
        media_type: Union[MediaType, str] = MediaType.HTML,
    ) -> None:
        """Handle the rendering of a given template into a bytes string.

        Args:
            template_name: Path-like name for the template to be rendered, e.g. "index.html".
            template_engine: The template engine class to use to render the response.
            status_code: A value for the response HTTP status code.
            context: A dictionary of key/value pairs to be passed to the temple engine's render method.
            background: A [BackgroundTask][starlite.datastructures.BackgroundTask] instance or
                [BackgroundTasks][starlite.datastructures.BackgroundTasks] to execute after the response is finished.
                Defaults to None.
            headers: A string keyed dictionary of response headers. Header keys are insensitive.
            cookies: A list of [Cookie][starlite.datastructures.Cookie] instances to be set under the response
                'Set-Cookie' header.
            encoding: Content encoding
            media_type: A string or member of the [MediaType][starlite.enums.MediaType] enum. If not set, try to infer
                the media type based on the template name. If this fails, fall back to `text/plain`.
        """
        if media_type == MediaType.JSON:  # we assume this is the default
            suffixes = PurePath(template_name).suffixes
            for suffix in suffixes:
                if _type := guess_type("name" + suffix)[0]:
                    media_type = _type
                    break
            else:
                media_type = MediaType.TEXT

        template = template_engine.get_template(template_name)
        super().__init__(
            background=background,
            content=template.render(**context),
            cookies=cookies,
            encoding=encoding,
            headers=headers,
            media_type=media_type,
            status_code=status_code,
        )



from typing import TYPE_CHECKING, Any, NamedTuple, Set

from pydantic.fields import Undefined

from starlite.constants import EXTRA_KEY_REQUIRED
from starlite.enums import ParamType

if TYPE_CHECKING:
    from pydantic.fields import FieldInfo


class ParameterDefinition(NamedTuple):
    """Tuple defining a kwarg representing a request parameter."""

    default_value: Any
    field_alias: str
    field_name: str
    is_required: bool
    is_sequence: bool
    param_type: ParamType


def create_parameter_definition(
    allow_none: bool, field_info: "FieldInfo", field_name: str, path_parameters: Set[str], is_sequence: bool
) -> ParameterDefinition:
    """Create a ParameterDefinition for the given pydantic FieldInfo instance and inserts it into the correct parameter
    set.

    Args:
        allow_none: Whether 'None' is an allowed value for the parameter.
        field_info: A pydantic field info.
        field_name: The field's name.
        path_parameters: A set of path parameter names.
        is_sequence: Whether the value of the parameter is a sequence.

    Returns:
        A ParameterDefinition tuple.
    """
    extra = field_info.extra
    is_required = extra.get(EXTRA_KEY_REQUIRED, True)
    default_value = field_info.default if field_info.default is not Undefined else None

    field_alias = extra.get(ParamType.QUERY) or field_name
    param_type = ParamType.QUERY

    if field_name in path_parameters:
        field_alias = field_name
        param_type = ParamType.PATH
    elif extra.get(ParamType.HEADER):
        field_alias = extra[ParamType.HEADER]
        param_type = ParamType.HEADER
    elif extra.get(ParamType.COOKIE):
        field_alias = extra[ParamType.COOKIE]
        param_type = ParamType.COOKIE

    return ParameterDefinition(
        param_type=param_type,
        field_name=field_name,
        field_alias=field_alias,
        default_value=default_value,
        is_required=is_required and (default_value is None and not allow_none),
        is_sequence=is_sequence,
    )


def merge_parameter_sets(first: Set[ParameterDefinition], second: Set[ParameterDefinition]) -> Set[ParameterDefinition]:
    """Given two sets of parameter definitions, coming from different dependencies for example, merge them into a single
    set.
    """
    result: Set[ParameterDefinition] = first.intersection(second)
    difference = first.symmetric_difference(second)
    for param in difference:
        # add the param if it's either required or no-other param in difference is the same but required
        if param.is_required or not any(p.field_alias == param.field_alias and p.is_required for p in difference):
            result.add(param)
    return result



from functools import lru_cache
from typing import TYPE_CHECKING, Any, NamedTuple, Optional, Union, cast
from urllib.parse import SplitResult, urlencode, urlsplit, urlunsplit

from starlite.datastructures import MultiDict
from starlite.parsers import parse_query_string
from starlite.types import Empty

if TYPE_CHECKING:
    from starlite.types import EmptyType, Scope

_DEFAULT_SCHEME_PORTS = {"http": 80, "https": 443, "ftp": 21, "ws": 80, "wss": 443}


class Address(NamedTuple):
    """Just a network address."""

    host: str
    """Address host."""
    port: int
    """Address port."""


def make_absolute_url(path: Union[str, "URL"], base: Union[str, "URL"]) -> str:
    """Create an absolute URL.

    Args:
        path: URL path to make absolute
        base: URL to use as a base

    Returns:
        A string representing the new, absolute URL
    """
    url = base if isinstance(base, URL) else URL(base)
    netloc = url.netloc
    path = url.path.rstrip("/") + str(path)
    return str(URL.from_components(scheme=url.scheme, netloc=netloc, path=path))


class URL:
    """Representation and modification utilities of a URL."""

    __slots__ = (
        "_query_params",
        "_parsed_url",
        "fragment",
        "hostname",
        "netloc",
        "password",
        "path",
        "port",
        "query",
        "scheme",
        "username",
    )

    _query_params: Union["EmptyType", "MultiDict"]
    _parsed_url: Optional[str]

    scheme: str
    """URL scheme."""
    netloc: str
    """Network location."""
    path: str
    """Hierarchical path."""
    fragment: str
    """Fragment component."""
    query: str
    """Query string."""
    username: Optional[str]
    """Username if specified."""
    password: Optional[str]
    """Password if specified."""
    port: Optional[int]
    """Port if specified."""
    hostname: Optional[str]
    """Hostname if specified."""

    @lru_cache  # type: ignore[misc]  # noqa: B019
    def __new__(cls, url: Union[str, SplitResult]) -> "URL":
        """Create a new instance.

        Args:
            url: url string or split result to represent.
        """
        instance = super().__new__(cls)
        instance._parsed_url = None

        if isinstance(url, str):
            result = urlsplit(url)
            instance._parsed_url = url
        else:
            result = url

        instance.scheme = result.scheme
        instance.netloc = result.netloc
        instance.path = result.path
        instance.fragment = result.fragment
        instance.query = result.query
        instance.username = result.username
        instance.password = result.password
        instance.port = result.port
        instance.hostname = result.hostname
        instance._query_params = Empty

        return instance

    @property
    def _url(self) -> str:
        if not self._parsed_url:
            self._parsed_url = urlunsplit(
                SplitResult(
                    scheme=self.scheme,
                    netloc=self.netloc,
                    path=self.path,
                    fragment=self.fragment,
                    query=self.query,
                )
            )
        return self._parsed_url

    @classmethod
    @lru_cache
    def from_components(
        cls,
        scheme: str = "",
        netloc: str = "",
        path: str = "",
        fragment: str = "",
        query: str = "",
    ) -> "URL":
        """Create a new URL from components.

        Args:
            scheme: URL scheme
            netloc: Network location
            path: Hierarchical path
            query: Query component
            fragment: Fragment identifier

        Returns:
            A new URL with the given components
        """
        return cls(  # type: ignore[no-any-return]
            SplitResult(
                scheme=scheme,
                netloc=netloc,
                path=path,
                fragment=fragment,
                query=query,
            )
        )

    @classmethod
    def from_scope(cls, scope: "Scope") -> "URL":
        """Construct a URL from a [Scope][starlite.types.Scope]

        Args:
            scope: A scope

        Returns:
            A URL
        """
        scheme = scope.get("scheme", "http")
        server = scope.get("server")
        path = scope.get("root_path", "") + scope["path"]
        query_string = scope.get("query_string", b"")

        # # we use iteration here because it's faster, and headers might not yet be cached
        # # in the scope
        host = ""
        for header_name, header_value in scope.get("headers", []):
            if header_name == b"host":
                host = header_value.decode("latin-1")
                break
        if server and not host:
            host, port = server
            default_port = _DEFAULT_SCHEME_PORTS[scheme]
            if port != default_port:
                host = f"{host}:{port}"

        return cls.from_components(
            scheme=scheme if server else "",
            query=query_string.decode(),
            netloc=host,
            path=path,
        )

    def with_replacements(
        self,
        scheme: str = "",
        netloc: str = "",
        path: str = "",
        query: Optional[Union[str, "MultiDict"]] = None,
        fragment: str = "",
    ) -> "URL":
        """Create a new URL, replacing the given components.

        Args:
            scheme: URL scheme
            netloc: Network location
            path: Hierarchical path
            query: Raw query string
            fragment: Fragment identifier

        Returns:
            A new URL with the given components replaced
        """
        if isinstance(query, MultiDict):
            query = urlencode(query=query)

        return URL.from_components(  # type: ignore[no-any-return]
            scheme=scheme or self.scheme,
            netloc=netloc or self.netloc,
            path=path or self.path,
            query=query or self.query,
            fragment=fragment or self.fragment,
        )

    @property
    def query_params(self) -> "MultiDict":
        """Query parameters of a URL as a [MultiDict][multidict.MultiDict]

        Returns:
            A [MultiDict][multidict.MultiDict] with query parameters

        Notes:
            - While the returned `MultiDict` is mutable, [URL][starlite.datastructures.URL]
                itself is *immutable*, therefore mutating the query parameters will not
                directly mutate the `URL`. If you want to modify query parameters, make
                modifications in the multidict and pass them back to
                [with_replacements][starlite.datastructures.URL.with_replacements]
        """
        if self._query_params is Empty:
            self._query_params = MultiDict(parse_query_string(query_string=self.query.encode()))
        return cast("MultiDict", self._query_params)

    def __str__(self) -> str:
        return self._url

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, (str, URL)):
            return str(self) == str(other)
        return NotImplemented  # type: ignore[unreachable]  # pragma: no cover

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._url!r})"



from .base import SessionMiddleware
from .cookie_backend import (
    CookieBackendConfig as SessionCookieConfig,  # backwards compatible export
)

__all__ = [
    "SessionMiddleware",
    "SessionCookieConfig",
]



from copy import copy
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, Extra, validate_arguments
from pydantic.fields import ModelField, Undefined

from starlite.constants import EXTRA_KEY_REQUIRED, EXTRA_KEY_VALUE_TYPE
from starlite.datastructures.provide import Provide
from starlite.exceptions import ImproperlyConfiguredException
from starlite.types import (
    Dependencies,
    Empty,
    EmptyType,
    ExceptionHandlersMap,
    Guard,
    Middleware,
    ParametersMap,
)
from starlite.types.composite_types import MaybePartial
from starlite.utils import AsyncCallable, Ref, get_name, normalize_path
from starlite.utils.helpers import unwrap_partial

if TYPE_CHECKING:
    from inspect import Signature

    from starlite.connection import ASGIConnection
    from starlite.controller import Controller
    from starlite.router import Router
    from starlite.signature import SignatureModel
    from starlite.types import AnyCallable

T = TypeVar("T", bound="BaseRouteHandler")


class ParameterConfig(BaseConfig):
    """Base config for `ModelField`"""

    extra = Extra.allow


class BaseRouteHandler(Generic[T]):
    """Base route handler.

    Serves as a subclass for all route handlers
    """

    fn: "Ref[MaybePartial[AnyCallable]]"
    signature: "Signature"

    __slots__ = (
        "_resolved_dependencies",
        "_resolved_guards",
        "_resolved_layered_parameters",
        "dependencies",
        "exception_handlers",
        "fn",
        "guards",
        "middleware",
        "name",
        "opt",
        "owner",
        "paths",
        "signature",
        "signature_model",
    )

    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(
        self,
        path: Union[Optional[str], Optional[List[str]]] = None,
        *,
        dependencies: Optional[Dependencies] = None,
        exception_handlers: Optional[ExceptionHandlersMap] = None,
        guards: Optional[List[Guard]] = None,
        middleware: Optional[List[Middleware]] = None,
        name: Optional[str] = None,
        opt: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize `HTTPRouteHandler`.

        Args:
            path: A path fragment for the route handler function or a list of path fragments. If not given defaults to '/'
            dependencies: A string keyed dictionary of dependency [Provider][starlite.datastructures.Provide] instances.
            exception_handlers: A dictionary that maps handler functions to status codes and/or exception types.
            guards: A list of [Guard][starlite.types.Guard] callables.
            middleware: A list of [Middleware][starlite.types.Middleware].
            name: A string identifying the route handler.
            opt: A string keyed dictionary of arbitrary values that can be accessed in [Guards][starlite.types.Guard] or
                wherever you have access to [Request][starlite.connection.request.Request] or [ASGI Scope][starlite.types.Scope].
            **kwargs: Any additional kwarg - will be set in the opt dictionary.
        """
        self._resolved_dependencies: Union[Dict[str, Provide], EmptyType] = Empty
        self._resolved_guards: Union[List[Guard], EmptyType] = Empty
        self._resolved_layered_parameters: Union[Dict[str, "ModelField"], EmptyType] = Empty
        self.dependencies = dependencies
        self.exception_handlers = exception_handlers
        self.guards = guards
        self.middleware = middleware
        self.name = name
        self.opt: Dict[str, Any] = opt or {}
        self.owner: Optional[Union["Controller", "Router"]] = None
        self.signature_model: Optional[Type["SignatureModel"]] = None
        self.paths = (
            {normalize_path(p) for p in path}
            if path and isinstance(path, list)
            else {normalize_path(path or "/")}  # type: ignore
        )
        self.opt.update(**kwargs)

    @property
    def handler_name(self) -> str:
        """Get the name of the handler function.

        Raises:
            ImproperlyConfiguredException: if handler fn is not set.

        Returns:
            Name of the handler function
        """
        fn = getattr(self, "fn", None)
        if not fn:
            raise ImproperlyConfiguredException("cannot access handler name before setting the handler function")
        return get_name(unwrap_partial(self.fn.value))

    @property
    def dependency_name_set(self) -> Set[str]:
        """Set of all dependency names provided in the handler's ownership layers."""
        layered_dependencies = (layer.dependencies or {} for layer in self.ownership_layers)
        return {name for layer in layered_dependencies for name in layer.keys()}

    @property
    def ownership_layers(self) -> List[Union[T, "Controller", "Router"]]:
        """Return the handler layers from the app down to the route handler.

        app -> ... -> route handler
        """
        layers = []

        cur: Any = self
        while cur:
            layers.append(cur)
            cur = cur.owner

        return list(reversed(layers))

    def resolve_layered_parameters(self) -> Dict[str, "ModelField"]:
        """Return all parameters declared above the handler, transforming them into pydantic ModelField instances."""
        if self._resolved_layered_parameters is Empty:
            self._resolved_layered_parameters = {}
            parameters: ParametersMap = {}
            for layer in self.ownership_layers:
                parameters.update(getattr(layer, "parameters", None) or {})

            for key, parameter in parameters.items():
                is_required = parameter.extra[EXTRA_KEY_REQUIRED]
                value_type = parameter.extra[EXTRA_KEY_VALUE_TYPE]
                if value_type is Undefined:
                    value_type = Any
                default_value = parameter.default if parameter.default is not Undefined else ...
                self._resolved_layered_parameters[key] = ModelField(
                    name=key,
                    type_=value_type,
                    field_info=parameter,
                    default=default_value,
                    model_config=ParameterConfig,
                    class_validators=None,
                    required=is_required,
                )
        return cast("Dict[str, ModelField]", self._resolved_layered_parameters)

    def resolve_guards(self) -> List[Guard]:
        """Return all guards in the handlers scope, starting from highest to current layer."""
        if self._resolved_guards is Empty:
            self._resolved_guards = []
            for layer in self.ownership_layers:
                self._resolved_guards.extend(layer.guards or [])
            self._resolved_guards = cast("List[Guard]", [AsyncCallable(guard) for guard in self._resolved_guards])
        return self._resolved_guards  # type:ignore

    def resolve_dependencies(self) -> Dict[str, Provide]:
        """Return all dependencies correlating to handler function's kwargs that exist in the handler's scope."""
        if not self.signature_model:
            raise RuntimeError("resolve_dependencies cannot be called before a signature model has been generated")
        if self._resolved_dependencies is Empty:
            self._resolved_dependencies = {}
            for layer in self.ownership_layers:
                for key, value in (layer.dependencies or {}).items():
                    self._validate_dependency_is_unique(
                        dependencies=self._resolved_dependencies, key=key, provider=value
                    )
                    self._resolved_dependencies[key] = value
        return cast("Dict[str, Provide]", self._resolved_dependencies)

    def resolve_middleware(self) -> List["Middleware"]:
        """Build the middleware stack for the RouteHandler and return it.

        The middlewares are added from top to bottom (app -> router -> controller -> route handler) and then reversed.
        """
        resolved_middleware = []
        for layer in self.ownership_layers:
            resolved_middleware.extend(layer.middleware or [])
        return list(reversed(resolved_middleware))

    def resolve_exception_handlers(self) -> ExceptionHandlersMap:
        """Resolve the exception_handlers by starting from the route handler and moving up.

        This method is memoized so the computation occurs only once.
        """
        resolved_exception_handlers = {}
        for layer in self.ownership_layers:
            resolved_exception_handlers.update(layer.exception_handlers or {})
        return resolved_exception_handlers

    def resolve_opts(self) -> None:
        """Build the route handler opt dictionary by going from top to bottom.

        If multiple layers define the same key, the value from the closest layer to the response handler will take
        precedence.
        """

        opt: Dict[str, Any] = {}
        for layer in self.ownership_layers:
            opt.update(layer.opt or {})

        self.opt = opt

    async def authorize_connection(self, connection: "ASGIConnection") -> None:
        """Ensure the connection is authorized by running all the route guards in scope."""
        for guard in self.resolve_guards():
            await guard(connection, copy(self))  # type: ignore

    @staticmethod
    def _validate_dependency_is_unique(dependencies: Dict[str, Provide], key: str, provider: Provide) -> None:
        """Validate that a given provider has not been already defined under a different key."""
        for dependency_key, value in dependencies.items():
            if provider == value:
                raise ImproperlyConfiguredException(
                    f"Provider for key {key} is already defined under the different key {dependency_key}. "
                    f"If you wish to override a provider, it must have the same key."
                )

    def _validate_handler_function(self) -> None:
        """Validate the route handler function once set by inspecting its return annotations."""
        if not getattr(self, "fn", None):
            raise ImproperlyConfiguredException("Cannot call _validate_handler_function without first setting self.fn")

    def __str__(self) -> str:
        """Return a unique identifier for the route handler.

        Returns:
            A string
        """
        target = unwrap_partial(self.fn.value)
        if not hasattr(target, "__qualname__"):
            target = type(target)
        return f"{target.__module__}.{target.__qualname__}"



from dataclasses import MISSING
from dataclasses import Field as DataclassField
from dataclasses import dataclass
from inspect import getmro
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    get_type_hints,
)

from pydantic import BaseModel, create_model
from pydantic.typing import is_classvar
from typing_extensions import TypedDict, get_args

from starlite.exceptions import ImproperlyConfiguredException
from starlite.types.builtin_types import NoneType
from starlite.utils.predicates import (
    is_class_and_subclass,
    is_dataclass_class_typeguard,
    is_typeddict_typeguard,
)

try:
    # python 3.9 changed these variable
    from typing import _UnionGenericAlias as GenericAlias  # type: ignore
except ImportError:  # pragma: no cover
    from typing import _GenericAlias as GenericAlias  # type: ignore

if TYPE_CHECKING:
    from typing import TypeAlias, Union  # noqa: F401  # nopycln: import

    from starlite.types.builtin_types import DataclassClass, TypedDictClass

T = TypeVar("T")

SupportedTypes: "TypeAlias" = "Union[DataclassClass, Type[BaseModel], TypedDictClass]"
"""Types that are supported by [`Partial`][starlite.types.partial.Partial]"""


class Partial(Generic[T]):
    """Type generation for PATCH routes.

    Partial is a special typing helper that takes a generic T, which must be a
    [`TypedDict`][typing.TypedDict], dataclass or pydantic model class, and
    returns to static type checkers a version of this T in which all fields -
    and nested fields - are optional.
    """

    _models: Dict[SupportedTypes, SupportedTypes] = {}

    def __class_getitem__(cls, item: Type[T]) -> Type[T]:
        """Take a pydantic model class, [`TypedDict`][typing.TypedDict] or a dataclass and return an all optional
        version of that class.

        Args:
            item: A pydantic model, [`TypedDict`][typing.TypedDict] or dataclass class.

        Returns:
            A pydantic model, [`TypedDict`][typing.TypedDict], or dataclass.
        """
        if item not in cls._models:
            if is_class_and_subclass(item, BaseModel):
                cls._create_partial_pydantic_model(item=item)
            elif is_dataclass_class_typeguard(item):
                cls._create_partial_dataclass(item=item)
            elif is_typeddict_typeguard(item):
                cls._create_partial_typeddict(item=item)
            else:
                raise ImproperlyConfiguredException(
                    "The type argument T passed to Partial[T] must be a `TypedDict`, dataclass or pydantic model class"
                )

        return cls._models[item]  # type:ignore[return-value]

    @classmethod
    def _create_partial_pydantic_model(cls, item: Type[BaseModel]) -> None:
        """Receives a pydantic model class and creates an all optional subclass of it.

        Args:
            item: A pydantic model class.
        """
        field_definitions: Dict[str, Tuple[Any, None]] = {}
        for field_name, field_type in get_type_hints(item).items():
            if is_classvar(field_type):
                continue
            if not isinstance(field_type, GenericAlias) or NoneType not in field_type.__args__:
                field_definitions[field_name] = (Optional[field_type], None)
            else:
                field_definitions[field_name] = (field_type, None)

        cls._models[item] = create_model(cls._create_partial_type_name(item), __base__=item, **field_definitions)  # type: ignore

    @classmethod
    def _create_partial_dataclass(cls, item: "DataclassClass") -> None:
        """Receives a dataclass class and creates an all optional subclass of it.

        Args:
            item: A dataclass class.
        """
        fields: Dict[str, DataclassField] = cls._create_optional_field_map(item)
        partial_type: "DataclassClass" = dataclass(
            type(cls._create_partial_type_name(item), (item,), {"__dataclass_fields__": fields})
        )
        annotated_ancestors = [a for a in getmro(partial_type) if hasattr(a, "__annotations__")]
        for ancestor in annotated_ancestors:
            for field_name, annotation in ancestor.__annotations__.items():
                if not isinstance(annotation, GenericAlias) or NoneType not in annotation.__args__:
                    partial_type.__annotations__[field_name] = Optional[annotation]
                else:
                    partial_type.__annotations__[field_name] = annotation

        cls._models[item] = partial_type

    @classmethod
    def _create_partial_typeddict(cls, item: "TypedDictClass") -> None:
        """Receives a typeddict class and creates a new type with all attributes `Optional`.

        Args:
            item: A [`TypedDict`][typing.TypeDict] class.
        """
        type_hints: Dict[str, Any] = {}
        for key_name, value_type in get_type_hints(item).items():
            if NoneType in get_args(value_type):
                type_hints[key_name] = value_type
                continue
            type_hints[key_name] = Optional[value_type]
        type_name = cls._create_partial_type_name(item)
        cls._models[item] = TypedDict(type_name, type_hints, total=False)  # type:ignore

    @staticmethod
    def _create_optional_field_map(item: "DataclassClass") -> Dict[str, DataclassField]:
        """Create a map of field name to optional dataclass Fields for a given dataclass.

        Args:
            item: A dataclass class.

        Returns:
            A map of field name to optional dataclass fields.
        """
        fields: Dict[str, DataclassField] = {}
        # https://github.com/python/typing/discussions/1056
        for field_name, dataclass_field in item.__dataclass_fields__.items():  # pyright:ignore
            if not isinstance(dataclass_field.type, GenericAlias) or NoneType not in dataclass_field.type.__args__:
                dataclass_field.type = Optional[dataclass_field.type]
            if dataclass_field.default_factory is MISSING:
                dataclass_field.default = None if dataclass_field.default is MISSING else dataclass_field.default
            fields[field_name] = dataclass_field
        return fields

    @staticmethod
    def _create_partial_type_name(item: SupportedTypes) -> str:
        return f"Partial{item.__name__}"



from typing import TYPE_CHECKING, Any, Optional

from starlite.constants import SCOPE_STATE_NAMESPACE
from starlite.types import Empty

if TYPE_CHECKING:
    from starlite.types import Scope, Serializer


def get_serializer_from_scope(scope: "Scope") -> Optional["Serializer"]:
    """Return a serializer given a scope object.

    Args:
        scope: The ASGI connection scope.

    Returns:
        A serializer function
    """
    route_handler = scope["route_handler"]
    if hasattr(route_handler, "resolve_response_class"):
        return route_handler.resolve_response_class().serializer  # pyright: ignore

    app = scope["app"]
    return app.response_class.serializer if app.response_class else None


def get_starlite_scope_state(scope: "Scope", key: str, default: Any = Empty) -> Any:
    """Get an internal value from connection scope state.

    Note:
        If called with a default value, this method behaves like to `dict.set_default()`, both setting the key in the
        namespace to the default value, and returning it.

        If called without a default value, the method behaves like `dict.get()`, returning `None` if the key does not
        exist.

    Args:
        scope: The connection scope.
        key: Key to get from internal namespace in scope state.
        default: Value set in internal namespace and returned if `key` doesn't exist.

    Returns:
        Value mapped to `key` in internal connection scope namespace.
    """
    namespace = scope["state"].setdefault(SCOPE_STATE_NAMESPACE, {})
    if default is Empty:
        return namespace.get(key)
    return namespace.setdefault(key, default)


def set_starlite_scope_state(scope: "Scope", key: str, value: Any) -> None:
    """Set an internal value in connection scope state.

    Args:
        scope: The connection scope.
        key: Key to set under internal namespace in scope state.
        value: Value for key.
    """
    scope["state"].setdefault(SCOPE_STATE_NAMESPACE, {})[key] = value



from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseConfig, BaseModel, validator
from pydantic_openapi_schema.v3_1_0 import SecurityRequirement

from starlite.connection import Request, WebSocket
from starlite.datastructures import CacheControlHeader, ETag, Provide
from starlite.plugins.base import PluginProtocol
from starlite.types import (
    AfterExceptionHookHandler,
    AfterRequestHookHandler,
    AfterResponseHookHandler,
    BeforeMessageSendHookHandler,
    BeforeRequestHookHandler,
    ControllerRouterHandler,
    ExceptionHandlersMap,
    Guard,
    LifeSpanHandler,
    LifeSpanHookHandler,
    Middleware,
    ParametersMap,
    ResponseCookies,
    ResponseHeadersMap,
    ResponseType,
    SingleOrList,
)

from . import AllowedHostsConfig
from .cache import CacheConfig
from .compression import CompressionConfig
from .cors import CORSConfig
from .csrf import CSRFConfig
from .logging import BaseLoggingConfig
from .openapi import OpenAPIConfig
from .static_files import StaticFilesConfig
from .template import TemplateConfig


class AppConfig(BaseModel):
    """The parameters provided to the `Starlite` app are used to instantiate an instance, and then the instance is
    passed to any callbacks registered to `on_app_init` in the order they are provided.

    The final attribute values are used to instantiate the application object.
    """

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    after_exception: SingleOrList[AfterExceptionHookHandler]
    """An application level [exception hook handler][starlite.types.AfterExceptionHookHandler] or list thereof.

    This hook is called after an exception occurs. In difference to exception handlers, it is not meant to return a
    response - only to process the exception (e.g. log it, send it to Sentry etc.).
    """
    after_request: Optional[AfterRequestHookHandler]
    """A sync or async function executed after the route handler function returned and the response object has been
    resolved.

    Receives the response object which may be any subclass of [Response][starlite.response.Response].
    """
    after_response: Optional[AfterResponseHookHandler]
    """A sync or async function called after the response has been awaited. It receives the.

    [Request][starlite.connection.Request] object and should not return any values.
    """
    after_shutdown: SingleOrList[LifeSpanHookHandler]
    """An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or list thereof.

    This hook is called during the ASGI shutdown, after all callables in the 'on_shutdown' list have been called.
    """
    after_startup: SingleOrList[LifeSpanHookHandler]
    """An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or list thereof.

    This hook is called during the ASGI startup, after all callables in the 'on_startup' list have been called.
    """
    allowed_hosts: Optional[Union[List[str], AllowedHostsConfig]]
    """If set enables the builtin allowed hosts middleware."""
    before_request: Optional[BeforeRequestHookHandler]
    """A sync or async function called immediately before calling the route handler. Receives the.

    [Request][starlite.connection.Request] instance and any non-`None` return value is used for the response, bypassing
    the route handler.
    """
    before_send: SingleOrList[BeforeMessageSendHookHandler]
    """An application level [before send hook handler][starlite.types.BeforeMessageSendHookHandler] or list thereof.

    This hook is called when the ASGI send function is called.
    """
    before_shutdown: SingleOrList[LifeSpanHookHandler]
    """An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or list thereof.

    This hook is called during the ASGI shutdown, before any callables in the 'on_shutdown' list have been called.
    """
    before_startup: SingleOrList[LifeSpanHookHandler]
    """An application level [life-span hook handler][starlite.types.LifeSpanHookHandler] or list thereof.

    This hook is called during the ASGI startup, before any callables in the 'on_startup' list have been called.
    """
    cache_config: CacheConfig
    """Configures caching behavior of the application."""
    cache_control: Optional[CacheControlHeader]
    """A `cache-control` header of type [CacheControlHeader][starlite.datastructures.CacheControlHeader] to add to route
    handlers of this app.

    Can be overridden by route handlers.
    """
    compression_config: Optional[CompressionConfig]
    """Configures compression behaviour of the application, this enabled a builtin or user defined Compression
    middleware.
    """
    cors_config: Optional[CORSConfig]
    """If set this enables the builtin CORS middleware."""
    csrf_config: Optional[CSRFConfig]
    """If set this enables the builtin CSRF middleware."""
    debug: bool
    """If `True`, app errors rendered as HTML with a stack trace."""
    dependencies: Dict[str, Provide]
    """A string keyed dictionary of dependency [Provider][starlite.datastructures.Provide] instances."""
    etag: Optional[ETag]
    """An `etag` header of type [ETag][starlite.datastructures.ETag] to add to route handlers of this app.

    Can be overridden by route handlers.
    """
    exception_handlers: ExceptionHandlersMap
    """A dictionary that maps handler functions to status codes and/or exception types."""
    guards: List[Guard]
    """A list of [Guard][starlite.types.Guard] callables."""
    logging_config: Optional[BaseLoggingConfig]
    """An instance of [BaseLoggingConfig][starlite.config.logging.BaseLoggingConfig] subclass."""
    middleware: List[Middleware]
    """A list of [Middleware][starlite.types.Middleware]."""
    on_shutdown: List[LifeSpanHandler]
    """A list of [LifeSpanHandler][starlite.types.LifeSpanHandler] called during application shutdown."""
    on_startup: List[LifeSpanHandler]
    """A list of [LifeSpanHandler][starlite.types.LifeSpanHandler] called during application startup."""
    openapi_config: Optional[OpenAPIConfig]
    """Defaults to [DEFAULT_OPENAPI_CONFIG][starlite.app.DEFAULT_OPENAPI_CONFIG]"""
    opt: Dict[str, Any]
    """A string keyed dictionary of arbitrary values that can be accessed in [Guards][starlite.types.Guard] or wherever
    you have access to [Request][starlite.connection.request.Request] or [ASGI Scope][starlite.types.Scope].

    Can be overridden by routers and router handlers.
    """
    parameters: ParametersMap
    """A mapping of [Parameter][starlite.params.Parameter] definitions available to all application paths."""
    plugins: List[PluginProtocol]
    """List of [PluginProtocol][starlite.plugins.base.PluginProtocol]."""
    request_class: Optional[Type[Request]]
    """An optional subclass of [Request][starlite.connection.request.Request] to use for http connections."""
    response_class: Optional[ResponseType]
    """A custom subclass of [starlite.response.Response] to be used as the app's default response."""
    response_cookies: ResponseCookies
    """A list of [Cookie](starlite.datastructures.Cookie] instances."""
    response_headers: ResponseHeadersMap
    """A string keyed dictionary mapping [ResponseHeader][starlite.datastructures.ResponseHeader] instances."""
    route_handlers: List[ControllerRouterHandler]
    """A required list of route handlers, which can include instances of [Router][starlite.router.Router], subclasses
    of.

    [Controller][starlite.controller.Controller] or any function decorated by the route handler decorators.
    """
    security: List[SecurityRequirement]
    """A list of dictionaries that will be added to the schema of all route handlers in the application. See.

    [SecurityRequirement][pydantic_openapi_schema.v3_1_0.security_requirement.SecurityRequirement] for details.
    """
    static_files_config: SingleOrList[StaticFilesConfig]
    """An instance or list of [StaticFilesConfig][starlite.config.StaticFilesConfig]."""
    tags: List[str]
    """A list of string tags that will be appended to the schema of all route handlers under the application."""
    template_config: Optional[TemplateConfig]
    """An instance of [TemplateConfig][starlite.config.TemplateConfig]."""
    websocket_class: Optional[Type[WebSocket]]
    """An optional subclass of [WebSocket][starlite.connection.websocket.WebSocket] to use for websocket connections."""

    @validator("allowed_hosts", always=True)
    def validate_allowed_hosts(  # pylint: disable=no-self-argument
        cls, value: Optional[Union[List[str], AllowedHostsConfig]]
    ) -> Optional[AllowedHostsConfig]:
        """Normalize the allowed hosts to be a config or None.

        Args:
            value: Optional a list of hosts or allowed hosts config

        Returns:
            Optional config.
        """
        if value:
            if isinstance(value, list):
                return AllowedHostsConfig(allowed_hosts=value)
            return value
        return None



from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Generic, Literal, Optional, Type, Union

from pydantic_openapi_schema.v3_1_0 import (
    Components,
    OAuthFlow,
    OAuthFlows,
    SecurityRequirement,
    SecurityScheme,
)

from starlite import Cookie, DefineMiddleware, Response
from starlite.contrib.jwt.jwt_token import Token
from starlite.contrib.jwt.middleware import (
    JWTAuthenticationMiddleware,
    JWTCookieAuthenticationMiddleware,
)
from starlite.enums import MediaType
from starlite.security.base import AbstractSecurityConfig, UserType
from starlite.status_codes import HTTP_201_CREATED


class JWTAuth(Generic[UserType], AbstractSecurityConfig[UserType, Token]):
    """JWT Authentication Configuration.

    This class is the main entry point to the library, and it includes methods to create the middleware, provide login
    functionality, and create OpenAPI documentation.
    """

    algorithm: str = "HS256"
    """Algorithm to use for JWT hashing."""
    auth_header: str = "Authorization"
    """Request header key from which to retrieve the token.

    E.g. 'Authorization' or 'X-Api-Key'.
    """
    default_token_expiration: timedelta = timedelta(days=1)
    """The default value for token expiration."""
    token_secret: str
    """Key with which to generate the token hash.

    Notes:
        - This value should be kept as a secret and the standard practice is to inject it into the environment.
    """
    openapi_security_scheme_name: str = "BearerToken"
    """The value to use for the OpenAPI security scheme and security requirements."""
    description: str = "JWT api-key authentication and authorization."
    """Description for the OpenAPI security scheme."""
    authentication_middleware_class: Type[JWTAuthenticationMiddleware] = JWTAuthenticationMiddleware
    """The authentication middleware class to use.

    Must inherit from [JWTAuthenticationMiddleware][starlite.contrib.jwt.JWTAuthenticationMiddleware]
    """

    @property
    def openapi_components(self) -> Components:
        """Create OpenAPI documentation for the JWT auth schema used.

        Returns:
            An [Components][pydantic_openapi_schema.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                self.openapi_security_scheme_name: SecurityScheme(
                    type="http",
                    scheme="Bearer",
                    name=self.auth_header,
                    bearerFormat="JWT",
                    description=self.description,
                )
            }
        )

    @property
    def security_requirement(self) -> SecurityRequirement:
        """Return OpenAPI 3.1.

        [SecurityRequirement][pydantic_openapi_schema.v3_1_0.security_requirement.SecurityRequirement]

        Returns:
            An OpenAPI 3.1 [SecurityRequirement][pydantic_openapi_schema.v3_1_0.security_requirement.SecurityRequirement] dictionary.
        """
        return {self.openapi_security_scheme_name: []}

    @property
    def middleware(self) -> DefineMiddleware:
        """Create `JWTAuthenticationMiddleware` wrapped in Starlite's `DefineMiddleware`.

        Returns:
            An instance of [DefineMiddleware][starlite.middleware.base.DefineMiddleware].
        """
        return DefineMiddleware(
            self.authentication_middleware_class,
            algorithm=self.algorithm,
            auth_header=self.auth_header,
            exclude=self.exclude,
            exclude_opt_key=self.exclude_opt_key,
            retrieve_user_handler=self.retrieve_user_handler,
            scopes=self.scopes,
            token_secret=self.token_secret,
        )

    def login(
        self,
        identifier: str,
        *,
        response_body: Optional[Any] = None,
        response_media_type: Union[str, MediaType] = MediaType.JSON,
        response_status_code: int = HTTP_201_CREATED,
        token_expiration: Optional[timedelta] = None,
        token_issuer: Optional[str] = None,
        token_audience: Optional[str] = None,
        token_unique_jwt_id: Optional[str] = None,
    ) -> Response[Any]:
        """Create a response with a JWT header. Calls the 'JWTAuth.store_token_handler' to persist the token 'sub'.

        Args:
            identifier: Unique identifier of the token subject. Usually this is a user ID or equivalent kind of value.
            response_body: An optional response body to send.
            response_media_type: An optional 'Content-Type'. Defaults to 'application/json'.
            response_status_code: An optional status code for the response. Defaults to '201 Created'.
            token_expiration: An optional timedelta for the token expiration.
            token_issuer: An optional value of the token 'iss' field.
            token_audience: An optional value for the token 'aud' field.
            token_unique_jwt_id: An optional value for the token 'jti' field.

        Returns:
            A [Response][starlite.response.Response] instance.
        """
        encoded_token = self.create_token(
            identifier=identifier,
            token_expiration=token_expiration,
            token_issuer=token_issuer,
            token_audience=token_audience,
            token_unique_jwt_id=token_unique_jwt_id,
        )
        return Response(
            content=response_body,
            headers={self.auth_header: self.format_auth_header(encoded_token)},
            media_type=response_media_type,
            status_code=response_status_code,
        )

    def create_token(
        self,
        identifier: str,
        token_expiration: Optional[timedelta] = None,
        token_issuer: Optional[str] = None,
        token_audience: Optional[str] = None,
        token_unique_jwt_id: Optional[str] = None,
    ) -> str:
        """Create a Token instance from the passed in parameters, persists and returns it.

        Args:
            identifier: Unique identifier of the token subject. Usually this is a user ID or equivalent kind of value.
            token_expiration: An optional timedelta for the token expiration.
            token_issuer: An optional value of the token 'iss' field.
            token_audience: An optional value for the token 'aud' field.
            token_unique_jwt_id: An optional value for the token 'jti' field.

        Returns:
            The created token.
        """
        token = Token(
            sub=identifier,
            exp=(datetime.now(timezone.utc) + (token_expiration or self.default_token_expiration)),
            iss=token_issuer,
            aud=token_audience,
            jti=token_unique_jwt_id,
        )
        return token.encode(secret=self.token_secret, algorithm=self.algorithm)

    def format_auth_header(self, encoded_token: str) -> str:
        """Format a token according to the specified OpenAPI scheme.

        Args:
            encoded_token: An encoded JWT token

        Returns:
            The encoded token formatted for the HTTP headers
        """
        security = self.openapi_components.securitySchemes.get(self.openapi_security_scheme_name, None)  # type: ignore
        return f"{security.scheme} {encoded_token}" if isinstance(security, SecurityScheme) else encoded_token


class JWTCookieAuth(Generic[UserType], JWTAuth[UserType]):
    """JWT Cookie Authentication Configuration.

    This class is an alternate entry point to the library, and it includes all the functionality of the `JWTAuth` class
    and adds support for passing JWT tokens `HttpOnly` cookies.
    """

    key: str = "token"
    """Key for the cookie."""
    path: str = "/"
    """Path fragment that must exist in the request url for the cookie to be valid.

    Defaults to '/'.
    """
    domain: Optional[str] = None
    """Domain for which the cookie is valid."""
    secure: Optional[bool] = None
    """Https is required for the cookie."""
    samesite: Literal["lax", "strict", "none"] = "lax"
    """Controls whether or not a cookie is sent with cross-site requests.

    Defaults to 'lax'.
    """
    description: str = "JWT cookie-based authentication and authorization."
    """Description for the OpenAPI security scheme."""
    authentication_middleware_class: Type[JWTCookieAuthenticationMiddleware] = JWTCookieAuthenticationMiddleware
    """The authentication middleware class to use.

    Must inherit from [JWTCookieAuthenticationMiddleware][starlite.contrib.jwt.JWTCookieAuthenticationMiddleware]
    """

    @property
    def openapi_components(self) -> Components:
        """Create OpenAPI documentation for the JWT Cookie auth scheme.

        Returns:
            An [Components][pydantic_openapi_schema.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                self.openapi_security_scheme_name: SecurityScheme(
                    type="http",
                    scheme="Bearer",
                    name=self.key,
                    security_scheme_in="cookie",
                    bearerFormat="JWT",
                    description=self.description,
                )
            }
        )

    @property
    def middleware(self) -> DefineMiddleware:
        """Create `JWTCookieAuthenticationMiddleware` wrapped in Starlite's `DefineMiddleware`.

        Returns:
            An instance of [DefineMiddleware][starlite.middleware.base.DefineMiddleware].
        """
        return DefineMiddleware(
            self.authentication_middleware_class,
            algorithm=self.algorithm,
            auth_cookie_key=self.key,
            auth_header=self.auth_header,
            exclude=self.exclude,
            exclude_opt_key=self.exclude_opt_key,
            retrieve_user_handler=self.retrieve_user_handler,
            scopes=self.scopes,
            token_secret=self.token_secret,
        )

    def login(
        self,
        identifier: str,
        *,
        response_body: Optional[Any] = None,
        response_media_type: Union[str, MediaType] = MediaType.JSON,
        response_status_code: int = HTTP_201_CREATED,
        token_expiration: Optional[timedelta] = None,
        token_issuer: Optional[str] = None,
        token_audience: Optional[str] = None,
        token_unique_jwt_id: Optional[str] = None,
    ) -> Response[Any]:
        """Create a response with a JWT header. Calls the 'JWTAuth.store_token_handler' to persist the token 'sub'.

        Args:
            identifier: Unique identifier of the token subject. Usually this is a user ID or equivalent kind of value.
            response_body: An optional response body to send.
            response_media_type: An optional 'Content-Type'. Defaults to 'application/json'.
            response_status_code: An optional status code for the response. Defaults to '201 Created'.
            token_expiration: An optional timedelta for the token expiration.
            token_issuer: An optional value of the token 'iss' field.
            token_audience: An optional value for the token 'aud' field.
            token_unique_jwt_id: An optional value for the token 'jti' field.

        Returns:
            A [Response][starlite.response.Response] instance.
        """
        encoded_token = self.create_token(
            identifier=identifier,
            token_expiration=token_expiration,
            token_issuer=token_issuer,
            token_audience=token_audience,
            token_unique_jwt_id=token_unique_jwt_id,
        )
        cookie = Cookie(
            key=self.key,
            path=self.path,
            httponly=True,
            value=self.format_auth_header(encoded_token),
            expires=int((datetime.now(timezone.utc) + (token_expiration or self.default_token_expiration)).timestamp()),
            secure=self.secure,
            samesite=self.samesite,
        )
        return Response(
            content=response_body,
            headers={self.auth_header: self.format_auth_header(encoded_token)},
            cookies=[cookie],
            media_type=response_media_type,
            status_code=response_status_code,
        )


class OAuth2PasswordBearerAuth(Generic[UserType], JWTCookieAuth[UserType]):
    """OAUTH2 Schema for Password Bearer Authentication.

    This class implements an OAUTH2 authentication flow entry point to the library, and it
    includes all the functionality of the `JWTAuth` class and adds
    support for passing JWT tokens `HttpOnly` cookies.

    `token_url` is the only additional argument that is required, and it should point at your login route
    """

    token_url: str
    """The URL for retrieving a new token."""
    oauth_scopes: Optional[Dict[str, str]] = None
    """Oauth Scopes available for the token."""
    description: str = "OAUTH2 password bearer authentication and authorization."
    """Description for the OpenAPI security scheme."""

    @property
    def oauth_flow(self) -> OAuthFlow:
        """Create an OpenAPI OAuth2 flow for the password bearer authentication scheme.

        Returns:
            An [OAuthFlow][pydantic_openapi_schema.v3_1_0.oauth_flow.OAuthFlow] instance.
        """
        return OAuthFlow(
            tokenUrl=self.token_url,
            scopes=self.oauth_scopes,
        )

    @property
    def openapi_components(self) -> Components:
        """Create OpenAPI documentation for the OAUTH2 Password bearer auth scheme.

        Returns:
            An [Components][pydantic_openapi_schema.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                self.openapi_security_scheme_name: SecurityScheme(
                    type="oauth2",
                    scheme="Bearer",
                    name=self.auth_header,
                    security_scheme_in="header",
                    flows=OAuthFlows(password=self.oauth_flow),  # pyright: reportGeneralTypeIssues=false
                    bearerFormat="JWT",
                    description=self.description,
                )
            }
        )



from typing import Any, Dict, Generic, Type

from pydantic_openapi_schema.v3_1_0 import (
    Components,
    SecurityRequirement,
    SecurityScheme,
)

from starlite.middleware.base import DefineMiddleware
from starlite.middleware.session.base import BaseBackendConfig, BaseSessionBackend
from starlite.security.base import AbstractSecurityConfig, UserType
from starlite.security.session_auth.middleware import (
    MiddlewareWrapper,
    SessionAuthMiddleware,
)


class SessionAuth(Generic[UserType], AbstractSecurityConfig[UserType, Dict[str, Any]]):
    """Session Based Security Backend."""

    session_backend_config: BaseBackendConfig
    """A session backend config."""

    authentication_middleware_class: Type[SessionAuthMiddleware] = SessionAuthMiddleware
    """The authentication middleware class to use.

    Must inherit from [SessionAuthMiddleware][starlite.security.session_auth.middleware.SessionAuthMiddleware]
    """

    @property
    def middleware(self) -> DefineMiddleware:
        """Use this property to insert the config into a middleware list on one of the application layers.

        Examples:
            ```python
            from typing import Any
            from os import urandom

            from starlite import Starlite, Request, get
            from starlite_session import SessionAuth


            async def retrieve_user_from_session(session: dict[str, Any]) -> Any:
                # implement logic here to retrieve a 'user' datum given the session dictionary
                ...


            session_auth_config = SessionAuth(
                secret=urandom(16), retrieve_user_handler=retrieve_user_from_session
            )


            @get("/")
            def my_handler(request: Request) -> None:
                ...


            app = Starlite(route_handlers=[my_handler], middleware=[session_auth_config.middleware])
            ```

        Returns:
            An instance of DefineMiddleware including 'self' as the config kwarg value.
        """
        return DefineMiddleware(MiddlewareWrapper, config=self)

    @property
    def session_backend(self) -> BaseSessionBackend:
        """Create a session backend.

        Returns:
            A subclass of [BaseSessionBackend][starlite.middleware.session.base.BaseSessionBackend]
        """
        return self.session_backend_config._backend_class(config=self.session_backend_config)

    @property
    def openapi_components(self) -> Components:
        """Create OpenAPI documentation for the Session Authentication schema used.

        Returns:
            An [Components][pydantic_openapi_schema.v3_1_0.components.Components] instance.
        """
        return Components(
            securitySchemes={
                "sessionCookie": SecurityScheme(
                    type="apiKey",
                    name=self.session_backend_config.key,
                    security_scheme_in="cookie",
                    description="Session cookie authentication.",
                )
            }
        )

    @property
    def security_requirement(self) -> SecurityRequirement:
        """Return OpenAPI 3.1.

        [SecurityRequirement][pydantic_openapi_schema.v3_1_0.security_requirement.SecurityRequirement] for the auth
        backend.

        Returns:
            An OpenAPI 3.1 [SecurityRequirement][pydantic_openapi_schema.v3_1_0.security_requirement.SecurityRequirement] dictionary.
        """
        return {"sessionCookie": []}



from inspect import Parameter, Signature
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generator,
    List,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from pydantic import BaseConfig, BaseModel, ValidationError, create_model
from pydantic.fields import FieldInfo, Undefined
from pydantic_factories import ModelFactory
from typing_extensions import get_args

from starlite.connection import Request
from starlite.enums import ScopeType
from starlite.exceptions import (
    ImproperlyConfiguredException,
    InternalServerException,
    ValidationException,
)
from starlite.plugins.base import PluginMapping, PluginProtocol, get_plugin_for_value
from starlite.utils import (
    is_dependency_field,
    is_optional_union,
    should_skip_dependency_validation,
)
from starlite.utils.helpers import unwrap_partial

if TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict

    from starlite import ASGIConnection
    from starlite.datastructures import URL
    from starlite.types import AnyCallable


UNDEFINED_SENTINELS = {Undefined, Signature.empty}
SKIP_NAMES = {"self", "cls"}
SKIP_VALIDATION_NAMES = {"request", "socket", "scope", "receive", "send"}


class SignatureModel(BaseModel):
    """Pydantic model representing a signature."""

    class Config(BaseConfig):
        copy_on_model_validation = "none"
        arbitrary_types_allowed = True

    dependency_name_set: ClassVar[Set[str]]
    field_plugin_mappings: ClassVar[Dict[str, PluginMapping]]
    return_annotation: ClassVar[Any]

    @classmethod
    def parse_values_from_connection_kwargs(cls, connection: "ASGIConnection", **kwargs: Any) -> Dict[str, Any]:
        """Given a dictionary of values extracted from the connection, create an instance of the given SignatureModel
        subclass and return the parsed values.

        This is not equivalent to calling the '.dict'  method of the pydantic model, because it doesn't convert nested
        values into dictionary, just extracts the data from the signature model
        """
        try:
            signature = cls(**kwargs)
            if signature.field_plugin_mappings:
                return {key: signature.resolve_field_value(key) for key in cls.__fields__}
            return {
                key: signature.__getattribute__(key)  # pylint: disable=unnecessary-dunder-call
                for key in cls.__fields__
            }
        except ValidationError as exc:
            raise cls.construct_exception(connection, exc) from exc

    def resolve_field_value(self, key: str) -> Any:
        """Given a field key, return value using plugin mapping, if available."""
        value = self.__getattribute__(key)  # pylint: disable=unnecessary-dunder-call
        mapping = self.field_plugin_mappings.get(key)
        return mapping.get_model_instance_for_value(value) if mapping else value

    @classmethod
    def construct_exception(
        cls, connection: "ASGIConnection", exc: ValidationError
    ) -> Union[InternalServerException, ValidationException]:
        """Distinguish between validation errors that arise from parameters and dependencies.

        If both parameter and dependency values are invalid, we raise the client error first.

        Args:
            connection: A Request or WebSocket
            exc: A ValidationError

        Returns:
            A ValidationException or InternalServerException
        """
        server_errors: List["ErrorDict"] = []
        client_errors: List["ErrorDict"] = []

        for error in exc.errors():
            if cls.is_server_error(error):
                server_errors.append(error)
            else:
                client_errors.append(error)

        method, url = cls.get_connection_method_and_url(connection)

        if client_errors:
            return ValidationException(detail=f"Validation failed for {method} {url}", extra=client_errors)

        return InternalServerException(detail=f"A dependency failed validation for {method} {url}", extra=server_errors)

    @classmethod
    def is_server_error(cls, error: "ErrorDict") -> bool:
        """Check whether given validation error is a server error."""
        return error["loc"][-1] in cls.dependency_name_set

    @staticmethod
    def get_connection_method_and_url(connection: "ASGIConnection") -> Tuple[str, "URL"]:
        """Extract method and URL from Request or WebSocket."""
        method = connection.method if isinstance(connection, Request) else ScopeType.WEBSOCKET
        return method, connection.url


class SignatureParameter:
    """Represents the parameters of a callable for purpose of signature model generation."""

    __slots__ = (
        "annotation",
        "default",
        "name",
        "optional",
    )

    annotation: Any
    default: Any
    name: str
    optional: bool

    def __init__(self, fn_name: str, parameter_name: str, parameter: Parameter) -> None:
        """Initialize SignatureParameter.

        Args:
            fn_name: Name of function.
            parameter_name: Name of parameter.
            parameter: inspect.Parameter
        """
        if parameter.annotation is Signature.empty:
            raise ImproperlyConfiguredException(
                f"Kwarg {parameter_name} of {fn_name} does not have a type annotation. If it "
                f"should receive any value, use the 'Any' type."
            )
        self.annotation = parameter.annotation
        self.default = parameter.default
        self.name = parameter_name
        self.optional = is_optional_union(parameter.annotation)

    @property
    def default_defined(self) -> bool:
        """Return a boolean, indicating if `self.default` is not one of the undefined sentinel types."""
        return self.default not in UNDEFINED_SENTINELS


class SignatureModelFactory:
    """Utility class for constructing the signature model and grouping associated state.

    Instance available at `SignatureModel.factory`.

    The following attributes are populated after the `model()` method has been called to generate
    the `SignatureModel` subclass.

    Attributes:
        field_plugin_mappings: Maps parameter name, to `PluginMapping` where a plugin has been applied.
        field_definitions:  Maps parameter name to the `(<type>, <default>)` tuple passed to `pydantic.create_model()`.
        defaults: Maps parameter name to default value, if one defined.
        dependency_name_set: The names of all known dependency parameters.
    """

    __slots__ = (
        "defaults",
        "dependency_name_set",
        "field_definitions",
        "field_plugin_mappings",
        "fn_module_name",
        "fn_name",
        "plugins",
        "signature",
    )

    def __init__(self, fn: "AnyCallable", plugins: List["PluginProtocol"], dependency_names: Set[str]) -> None:
        """Initialise `SignatureModelFactory`

        Args:
            fn: A callable
            plugins: A list of plugins
            dependency_names: Dependency names
        """
        if fn is None:
            raise ImproperlyConfiguredException("Parameter `fn` to `SignatureModelFactory` cannot be `None`.")

        fn = unwrap_partial(fn)
        self.signature = Signature.from_callable(fn)
        self.fn_name = getattr(fn, "__name__", "anonymous")
        self.fn_module_name = getattr(fn, "__module__", "pydantic.main")
        self.plugins = plugins
        self.field_plugin_mappings: Dict[str, PluginMapping] = {}
        self.field_definitions: Dict[str, Any] = {}
        self.defaults: Dict[str, Any] = {}
        self.dependency_name_set = dependency_names

    def check_for_unprovided_dependency(self, parameter: SignatureParameter) -> None:
        """Where a dependency has been explicitly marked using the `Dependency` function, it is a configuration error if
        that dependency has been defined without a default value, and it hasn't been provided to the handler.

        Args:
            parameter  SignatureParameter

        Raises:
            `ImproperlyConfiguredException`
        """
        if parameter.optional:
            return
        if not is_dependency_field(parameter.default):
            return
        field_info: FieldInfo = parameter.default
        if field_info.default is not Undefined:
            return
        if parameter.name not in self.dependency_name_set:
            raise ImproperlyConfiguredException(
                f"Explicit dependency '{parameter.name}' for '{self.fn_name}' has no default value, "
                f"or provided dependency."
            )

    def collect_dependency_names(self, parameter: SignatureParameter) -> None:
        """Add parameter name of dependencies declared using `Dependency()` function to the set of all dependency names.

        Args:
            parameter: SignatureParameter
        """
        if is_dependency_field(parameter.default):
            self.dependency_name_set.add(parameter.name)

    def set_field_default(self, parameter: SignatureParameter) -> None:
        """If `parameter` has defined default, map it to the parameter name in `self.defaults`.

        Args:
            parameter: SignatureParameter
        """
        if parameter.default_defined:
            self.defaults[parameter.name] = parameter.default

    def get_type_annotation_from_plugin(self, parameter: SignatureParameter, plugin: PluginProtocol) -> Any:
        """Use plugin declared for parameter annotation type to generate a pydantic model.

        Args:
            parameter:  SignatureParameter
            plugin: PluginProtocol

        Returns:
            A pydantic model to be used as a type
        """
        type_args = get_args(parameter.annotation)
        type_value = type_args[0] if type_args else parameter.annotation
        self.field_plugin_mappings[parameter.name] = PluginMapping(plugin=plugin, model_class=type_value)
        pydantic_model = plugin.to_pydantic_model_class(model_class=type_value, parameter_name=parameter.name)
        if type_args:
            return List[pydantic_model]  # type:ignore[valid-type]
        return pydantic_model

    @staticmethod
    def create_field_definition_from_parameter(parameter: SignatureParameter) -> Tuple[Any, Any]:
        """Construct an `(<annotation>, <default>)` tuple, appropriate for `pydantic.create_model()`.

        Args:
            parameter: SignatureParameter

        Returns:
            tuple[Any, Any]
        """
        if parameter.default_defined:
            return parameter.annotation, parameter.default
        if not parameter.optional:
            return parameter.annotation, ...
        return parameter.annotation, None

    @property
    def signature_parameters(self) -> Generator[SignatureParameter, None, None]:
        """Create a generator of `SignatureParameters` to be included in the `SignatureModel` by iterating over the
        parameters of the function signature.

        Returns:
            Generator[SignatureParameter, None, None]
        """
        for name, parameter in filter(lambda x: x[0] not in SKIP_NAMES, self.signature.parameters.items()):
            yield SignatureParameter(self.fn_name, name, parameter)

    @staticmethod
    def should_skip_parameter_validation(parameter: SignatureParameter) -> bool:
        """Identify dependencies for which provided values should not be validated.

        Args:
            parameter (SignatureParameter): A parameter to be added to the signature model.

        Returns:
            A boolean indicating whether injected values for this parameter should not be validated.
        """
        return parameter.name in SKIP_VALIDATION_NAMES or should_skip_dependency_validation(parameter.default)

    def create_signature_model(self) -> Type[SignatureModel]:
        """Construct a `SignatureModel` type that represents the signature of `self.fn`

        Returns:
            type[SignatureModel]
        """
        try:
            for parameter in self.signature_parameters:
                self.check_for_unprovided_dependency(parameter)
                self.collect_dependency_names(parameter)
                self.set_field_default(parameter)
                if self.should_skip_parameter_validation(parameter):
                    if is_dependency_field(parameter.default):
                        self.field_definitions[parameter.name] = (Any, parameter.default.default)
                    else:
                        self.field_definitions[parameter.name] = (Any, ...)
                    continue
                if ModelFactory.is_constrained_field(parameter.default):
                    self.field_definitions[parameter.name] = (parameter.default, ...)
                    continue
                plugin = get_plugin_for_value(value=parameter.annotation, plugins=self.plugins)
                if plugin:
                    parameter.annotation = self.get_type_annotation_from_plugin(parameter, plugin)
                self.field_definitions[parameter.name] = self.create_field_definition_from_parameter(parameter)
            model: Type[SignatureModel] = create_model(
                self.fn_name + "_signature_model",
                __base__=SignatureModel,
                __module__=self.fn_module_name,
                **self.field_definitions,
            )
            model.return_annotation = self.signature.return_annotation
            model.field_plugin_mappings = self.field_plugin_mappings
            model.dependency_name_set = self.dependency_name_set
            return model
        except TypeError as e:
            raise ImproperlyConfiguredException(f"Error creating signature model for '{self.fn_name}': '{e}'") from e


def get_signature_model(value: Any) -> Type[SignatureModel]:
    """Retrieve and validate the signature model from a provider or handler."""
    try:
        return cast("Type[SignatureModel]", value.signature_model)
    except AttributeError as e:  # pragma: no cover
        raise ImproperlyConfiguredException(f"The 'signature_model' attribute for {value} is not set") from e



from inspect import getmro
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, cast

from pydantic import BaseModel

from starlite.status_codes import HTTP_500_INTERNAL_SERVER_ERROR

if TYPE_CHECKING:
    from typing import Type

    from starlite.response import Response
    from starlite.types import ExceptionHandler, ExceptionHandlersMap


def get_exception_handler(exception_handlers: "ExceptionHandlersMap", exc: Exception) -> Optional["ExceptionHandler"]:
    """Given a dictionary that maps exceptions and status codes to handler functions, and an exception, returns the
    appropriate handler if existing.

    Status codes are given preference over exception type.

    If no status code match exists, each class in the MRO of the exception type is checked and
    the first matching handler is returned.

    Finally, if a `500` handler is registered, it will be returned for any exception that isn't a
    subclass of `HTTPException`.

    Args:
        exception_handlers: Mapping of status codes and exception types to handlers.
        exc: Exception Instance to be resolved to a handler.

    Returns:
        Optional exception handler callable.
    """
    if not exception_handlers:
        return None
    status_code: Optional[int] = getattr(exc, "status_code", None)
    if status_code in exception_handlers:
        return exception_handlers[status_code]  # pyright: ignore
    for cls in getmro(type(exc)):
        if cls in exception_handlers:
            return exception_handlers[cast("Type[Exception]", cls)]
    if not hasattr(exc, "status_code") and HTTP_500_INTERNAL_SERVER_ERROR in exception_handlers:
        return exception_handlers[HTTP_500_INTERNAL_SERVER_ERROR]
    return None


class ExceptionResponseContent(BaseModel):
    """Represent the contents of an exception-response."""

    status_code: int
    """Exception status code."""
    detail: str
    """Exception details or message."""
    headers: Optional[Dict[str, str]] = None
    """Headers to attach to the response."""
    extra: Optional[Union[Dict[str, Any], List[Any]]] = None
    """An extra mapping to attach to the exception."""

    def to_response(self) -> "Response":
        """Create a response from the model attributes.

        Returns:
            A response instance.
        """
        from starlite.response import Response

        return Response(
            content=self.dict(exclude_none=True, exclude={"headers"}),
            headers=self.headers,
            status_code=self.status_code,
        )


def create_exception_response(exc: Exception) -> "Response":
    """Construct a response from an exception.

    Notes:
    - For instances of [HTTPException][starlite.exceptions.HTTPException] or other exception classes that have a
        `status_code` attribute (e.g. Starlette exceptions), the status code is drawn from the exception, otherwise
        response status is `HTTP_500_INTERNAL_SERVER_ERROR`.

    Args:
        exc: An exception.

    Returns:
        Response: HTTP response constructed from exception details.
    """
    content = ExceptionResponseContent(
        status_code=getattr(exc, "status_code", HTTP_500_INTERNAL_SERVER_ERROR),
        detail=getattr(exc, "detail", repr(exc)),
        headers=getattr(exc, "headers", None),
        extra=getattr(exc, "extra", None),
    )
    return content.to_response()



from typing import Dict, Optional, Type

from aiomcache import Client as MemcacheClient

from starlite.middleware.session.base import ServerSideBackend, ServerSideSessionConfig
from starlite.utils.deprecation import deprecated


class MemcachedBackend(ServerSideBackend["MemcachedBackendConfig"]):
    """Session backend to store data in memcached."""

    __slots__ = ("memcached",)

    def __init__(self, config: "MemcachedBackendConfig") -> None:
        """Initialize `MemcachedBackend`

        Args:
            config: A `MemcachedBackendConfig` instance

        Notes:
            - Requires `aiomcache`. Install with `pip install starlite[memcached]`
        """
        super().__init__(config=config)
        self.memcached = config.memcached

    def _id_to_storage_key(self, session_id: str) -> bytes:
        return f"{self.config.key_prefix}:{session_id}".encode()

    async def get(self, session_id: str) -> Optional[bytes]:
        """Retrieve data associated with `session_id` from memcached.

        Args:
            session_id: The session-ID

        Returns:
            The session data, if existing, otherwise `None`.
        """
        return await self.memcached.get(key=self._id_to_storage_key(session_id))

    async def set(self, session_id: str, data: bytes) -> None:
        """Store `data` in memcached under `<prefix>:<session_id>`. If there is already data associated with
        `session_id`, replace it with `data` and reset its expiry time.

        Args:
            session_id: The session-ID
            data: Serialized session data

        Returns:
            None
        """
        await self.memcached.set(key=self._id_to_storage_key(session_id), value=data, exptime=self.config.max_age)

    async def delete(self, session_id: str) -> None:
        """Delete the data associated with `session_id`. Fail silently if no such session-ID exists.

        Args:
            session_id: The session-ID

        Returns:
            None
        """
        await self.memcached.delete(self._id_to_storage_key(session_id))

    @deprecated(
        "1.43.0",
        info="This functionality is not natively supported by memcached. Use the redis backend instead if you require it.",
    )
    async def delete_all(self) -> None:  # pragma: no cover
        """Delete all data stored within this backend.

        Returns:
            None

        Notes:
            This has poor performance since memcached does not offer utilities to
            properly scan or match keys by prefix.

        !!! important "Deprecated since 1.43.0"
            This method is deprecated since 1.43.0. If you need this functionality,
            consider using the redis backend instead.
        """
        stats: Dict[bytes, Optional[bytes]] = await self.memcached.stats(b"items")
        for key, raw_val in stats.items():
            if not raw_val:
                continue
            _, slab, field = key.split(b":")
            val = int(raw_val)
            if field != b"number" or val == 0:
                continue
            item_request: Dict[bytes, Optional[bytes]] = await self.memcached.stats(  # type: ignore[call-arg]
                b"cachedump", slab, str(val + 10).encode()  # pyright: ignore
            )
            for keys in item_request:
                await self.memcached.delete(keys)


class MemcachedBackendConfig(ServerSideSessionConfig):
    """Configuration for `MemcachedBackend`"""

    _backend_class: Type[MemcachedBackend] = MemcachedBackend

    memcached: MemcacheClient
    """An `aiomcache.Client` instance."""
    key_prefix: str = "STARLITE_SESSION"
    """Prefix to store data under after the schema of `<prefix>:<session-ID>`"""



from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from starlite.datastructures import Cookie, ETag
from starlite.enums import MediaType, OpenAPIMediaType
from starlite.exceptions import ImproperlyConfiguredException
from starlite.status_codes import (
    HTTP_200_OK,
    HTTP_204_NO_CONTENT,
    HTTP_304_NOT_MODIFIED,
)
from starlite.utils.helpers import get_enum_string_value
from starlite.utils.serialization import (
    DEFAULT_TYPE_ENCODERS,
    default_serializer,
    encode_json,
    encode_msgpack,
)

if TYPE_CHECKING:

    from starlite.datastructures import BackgroundTask, BackgroundTasks
    from starlite.types import (
        HTTPResponseBodyEvent,
        HTTPResponseStartEvent,
        Receive,
        ResponseCookies,
        Scope,
        Send,
    )

T = TypeVar("T")


class Response(Generic[T]):
    """Base Starlite HTTP response class, used as the basis for all other response classes."""

    __slots__ = (
        "background",
        "body",
        "cookies",
        "encoding",
        "headers",
        "is_head_response",
        "is_text_response",
        "media_type",
        "status_allows_body",
        "status_code",
        "raw_headers",
    )

    type_encoders: Dict[Any, Callable[[Any], Any]] = DEFAULT_TYPE_ENCODERS
    """Encoders for non-natively supported types."""

    def __init__(
        self,
        content: T,
        *,
        status_code: int = HTTP_200_OK,
        media_type: Union[MediaType, "OpenAPIMediaType", str] = MediaType.JSON,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
        encoding: str = "utf-8",
        is_head_response: bool = False,
    ) -> None:
        """Initialize the response.

        Args:
            content: A value for the response body that will be rendered into bytes string.
            status_code: An HTTP status code.
            media_type: A value for the response 'Content-Type' header.
            background: A [BackgroundTask][starlite.datastructures.BackgroundTask] instance or
                [BackgroundTasks][starlite.datastructures.BackgroundTasks] to execute after the response is finished.
                Defaults to None.
            headers: A string keyed dictionary of response headers. Header keys are insensitive.
            cookies: A list of [Cookie][starlite.datastructures.Cookie] instances to be set under
                the response 'Set-Cookie' header.
            encoding: The encoding to be used for the response headers.
            is_head_response: Whether the response should send only the headers ("head" request) or also the content.
        """
        self.background = background
        self.cookies = cookies or []
        self.encoding = encoding
        self.headers = headers or {}
        self.is_head_response = is_head_response
        self.media_type = get_enum_string_value(media_type)
        self.status_allows_body = not (
            status_code in {HTTP_204_NO_CONTENT, HTTP_304_NOT_MODIFIED} or status_code < HTTP_200_OK
        )
        self.status_code = status_code

        if not self.status_allows_body or is_head_response:
            if content:
                raise ImproperlyConfiguredException(
                    "response content is not supported for HEAD responses and responses with a status code "
                    "that does not allow content (304, 204, < 200)"
                )
            self.body = b""
        else:
            self.body = content if isinstance(content, bytes) else self.render(content)

        self.headers.setdefault(
            "content-type",
            f"{self.media_type}; charset={self.encoding}" if self.media_type.startswith("text/") else self.media_type,
        )
        self.raw_headers: List[Tuple[bytes, bytes]] = []

    @overload
    def set_cookie(self, /, cookie: Cookie) -> None:
        ...

    @overload
    def set_cookie(
        self,
        key: str,
        value: Optional[str] = None,
        max_age: Optional[int] = None,
        expires: Optional[int] = None,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
    ) -> None:
        ...

    def set_cookie(  # type: ignore[misc]
        self,
        key: Union[str, Cookie],
        value: Optional[str] = None,
        max_age: Optional[int] = None,
        expires: Optional[int] = None,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
    ) -> None:
        """Set a cookie on the response. If passed a [Cookie][starlite.Cookie] instance, keyword arguments will be
        ignored.

        Args:
            key: Key for the cookie or a [Cookie][starlite.Cookie] instance.
            value: Value for the cookie, if none given defaults to empty string.
            max_age: Maximal age of the cookie before its invalidated.
            expires: Expiration date as unix MS timestamp.
            path: Path fragment that must exist in the request url for the cookie to be valid. Defaults to '/'.
            domain: Domain for which the cookie is valid.
            secure: Https is required for the cookie.
            httponly: Forbids javascript to access the cookie via 'Document.cookie'.
            samesite: Controls whether a cookie is sent with cross-site requests. Defaults to 'lax'.

        Returns:
            None.
        """
        if not isinstance(key, Cookie):
            key = Cookie(
                domain=domain,
                expires=expires,
                httponly=httponly,
                key=key,
                max_age=max_age,
                path=path,
                samesite=samesite,
                secure=secure,
                value=value,
            )
        self.cookies.append(key)

    def set_header(self, key: str, value: str) -> None:
        """Set a header on the response.

        Args:
            key: Header key.
            value: Header value.

        Returns:
            None.
        """
        self.headers[key] = value

    def set_etag(self, etag: Union[str, "ETag"]) -> None:
        """Set an etag header.

        Args:
            etag: An etag value.

        Returns:
            None
        """
        self.headers["etag"] = etag.to_header() if isinstance(etag, ETag) else etag

    def delete_cookie(
        self,
        key: str,
        path: str = "/",
        domain: Optional[str] = None,
    ) -> None:
        """Delete a cookie.

        Args:
            key: Key of the cookie.
            path: Path of the cookie.
            domain: Domain of the cookie.

        Returns:
            None.
        """
        cookie = Cookie(key=key, path=path, domain=domain, expires=0, max_age=0)
        self.cookies = [c for c in self.cookies if c != cookie]
        self.cookies.append(cookie)

    @classmethod
    def serializer(cls, value: Any) -> Any:
        """Transform non-natively supported types into supported types using `Response.type_encoders`.

        Should raise `TypeError` if a type cannot be transformed into a supported type
        """

        return default_serializer(value, cls.type_encoders)

    def render(self, content: Any) -> bytes:
        """Handle the rendering of content T into a bytes string.

        Args:
            content: A value for the response body that will be rendered into bytes string.

        Returns:
            An encoded bytes string
        """
        try:
            if self.media_type.startswith("text/"):
                return content.encode(self.encoding) if content else b""  # type: ignore
            if self.media_type == MediaType.MESSAGEPACK:
                return encode_msgpack(content, self.serializer)
            return encode_json(content, self.serializer)
        except (AttributeError, ValueError, TypeError) as e:
            raise ImproperlyConfiguredException("Unable to serialize response content") from e

    @property
    def content_length(self) -> int:
        """Content length of the response if applicable.

        Returns:
            The content length of the body (e.g. for use in a "Content-Length" header).
            If the response does not have a body, this value is `None`
        """
        if self.status_allows_body:
            return len(self.body)
        return 0

    def encode_headers(self) -> List[Tuple[bytes, bytes]]:
        """Encode the response headers as a list of byte tuples.

        Notes:
            - A 'Content-Length' header will be added if appropriate and not provided by the user.

        Returns:
            A list of tuples containing the headers and cookies of the request in a format ready for ASGI transmission.
        """

        return list(
            chain(
                ((k.lower().encode("latin-1"), str(v).encode("latin-1")) for k, v in self.headers.items()),
                (cookie.to_encoded_header() for cookie in self.cookies),
                self.raw_headers,
            )
        )

    async def after_response(self) -> None:
        """Execute after the response is sent.

        Returns:
            None
        """
        if self.background is not None:
            await self.background()

    async def start_response(self, send: "Send") -> None:
        """Emit the start event of the response. This event includes the headers and status codes.

        Args:
            send: The ASGI send function.

        Returns:
            None
        """

        encoded_headers = self.encode_headers()

        content_length = self.content_length
        if "content-length" not in self.headers and content_length:
            encoded_headers.append((b"content-length", str(content_length).encode("latin-1")))

        event: "HTTPResponseStartEvent" = {
            "type": "http.response.start",
            "status": self.status_code,
            "headers": encoded_headers,
        }

        await send(event)

    async def send_body(self, send: "Send", receive: "Receive") -> None:  # pylint: disable=unused-argument
        """Emit the response body.

        Args:
            send: The ASGI send function.
            receive: The ASGI receive function.

        Notes:
            - Response subclasses should customize this method if there is a need to customize sending data.

        Returns:
            None
        """
        event: "HTTPResponseBodyEvent" = {"type": "http.response.body", "body": self.body, "more_body": False}
        await send(event)

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """ASGI callable of the `Response`.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        await self.start_response(send=send)

        if self.is_head_response:
            event: "HTTPResponseBodyEvent" = {"type": "http.response.body", "body": b"", "more_body": False}
            await send(event)
        else:
            await self.send_body(send=send, receive=receive)

        await self.after_response()



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, Optional, Type, TypeVar, Union, cast

import sqlalchemy as sa
from anyio.to_thread import run_sync
from pydantic import validator
from sqlalchemy.ext.asyncio import AsyncSession as AsyncSASession
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import Session as SASession
from sqlalchemy.orm import declarative_mixin, registry

from starlite.middleware.session.base import ServerSideBackend, ServerSideSessionConfig
from starlite.plugins.sql_alchemy import SQLAlchemyPlugin

if TYPE_CHECKING:
    from sqlalchemy.sql import Select
    from sqlalchemy.sql.elements import BooleanClauseList

    from starlite.plugins.sql_alchemy import SQLAlchemyConfig as SQLAlchemyPluginConfig


AnySASession = Union[SASession, AsyncSASession]
AnySASessionT = TypeVar("AnySASessionT", bound=AnySASession)
SessionModelT = TypeVar("SessionModelT", bound="SessionModelMixin")


@declarative_mixin
class SessionModelMixin:
    """Mixin for session storage."""

    session_id: Mapped[str] = sa.Column(sa.String, nullable=False, unique=True, index=True)  # pyright: ignore
    data: Mapped[bytes] = sa.Column(sa.LargeBinary, nullable=False)  # pyright: ignore
    expires: Mapped[datetime] = sa.Column(sa.DateTime, nullable=False)  # pyright: ignore

    @hybrid_property
    def expired(self) -> bool:  # pyright: ignore
        """Boolean indicating if the session has expired."""
        return datetime.utcnow().replace(tzinfo=None) > self.expires

    @expired.expression  # type: ignore[no-redef]
    def expired(cls) -> "BooleanClauseList":  # pylint: disable=no-self-argument
        """SQL-Expression to check if the session has expired."""
        return datetime.utcnow().replace(tzinfo=None) > cls.expires  # pyright: ignore


class SessionModel(SessionModelMixin):
    """Session storage model."""

    __tablename__ = "session"
    id: Mapped[int] = sa.Column(sa.Integer, primary_key=True)  # pyright: ignore


def create_session_model(base: Type[Any], table_name: str = "session") -> Type[SessionModelMixin]:
    """Dynamically generate a session storage model and register it with the declarative base.

    Args:
        base: SQLAlchemy declarative base
        table_name: Alternative table name

    Returns:
        A mapped model subclassing `base` and `SessionModelMixin`
    """

    class Model(base, SessionModelMixin):  # type: ignore[valid-type,misc]
        __tablename__ = table_name
        id: Mapped[int] = sa.Column(sa.Integer, primary_key=True)  # pyright: ignore

    return Model


def register_session_model(base: Union[registry, Any], model: Type[SessionModelT]) -> Type[SessionModelT]:
    """Map and register a pre-existing model subclassing `SessionModelMixin` with a declarative base or registry.

    Args:
        base: Either a `orm.registry` or `DeclarativeBase`
        model: SQLAlchemy model to register

    Returns:
        A mapped model subclassing `SessionModelMixin`, and registered in `registry`
    """
    registry_ = base.registry if not isinstance(base, registry) else base
    return cast("Type[SessionModelT]", registry_.map_declaratively(model))


class BaseSQLAlchemyBackend(Generic[AnySASessionT], ServerSideBackend["SQLAlchemyBackendConfig"], ABC):
    """Session backend to store data in a database with SQLAlchemy. Works with both sync and async engines.

    Notes:
        - Requires `sqlalchemy` which needs to be installed separately, and a configured
        [SQLAlchemyPlugin][starlite.plugins.sql_alchemy.SQLAlchemyPlugin].
    """

    __slots__ = ("_model", "_session_maker")

    def __init__(self, config: "SQLAlchemyBackendConfig") -> None:
        """Initialize `BaseSQLAlchemyBackend`.

        Args:
            config: An instance of `SQLAlchemyBackendConfig`
        """
        super().__init__(config=config)
        self._model = config.model
        self._session_maker = cast("SQLAlchemyPluginConfig", config.plugin._config).session_maker

    def _create_sa_session(self) -> AnySASessionT:
        return cast("AnySASessionT", self._session_maker())

    def _select_session_obj(self, session_id: str) -> "Select":
        return sa.select(self._model).where(self._model.session_id == session_id)

    def _update_session_expiry(self, session_obj: SessionModelMixin) -> None:
        session_obj.expires = datetime.utcnow().replace(tzinfo=None) + timedelta(seconds=self.config.max_age)

    @abstractmethod
    async def delete_expired(self) -> None:
        """Delete all expired sessions from the database."""


class AsyncSQLAlchemyBackend(BaseSQLAlchemyBackend[AsyncSASession]):
    """Asynchronous SQLAlchemy backend."""

    async def _get_session_obj(self, *, sa_session: AsyncSASession, session_id: str) -> Optional[SessionModelMixin]:
        result = await sa_session.scalars(self._select_session_obj(session_id))
        return result.one_or_none()

    async def get(self, session_id: str) -> Optional[bytes]:
        """Retrieve data associated with `session_id`.

        Args:
            session_id: The session-ID

        Returns:
            The session data, if existing, otherwise `None`.
        """
        async with self._create_sa_session() as sa_session:
            session_obj = await self._get_session_obj(sa_session=sa_session, session_id=session_id)
            if session_obj:
                if not session_obj.expired:  # type: ignore[truthy-function]
                    self._update_session_expiry(session_obj)  # type: ignore[unreachable]
                    await sa_session.commit()
                    return session_obj.data
                await sa_session.delete(session_obj)
                await sa_session.commit()
        return None

    async def set(self, session_id: str, data: bytes) -> None:
        """Store `data` under the `session_id` for later retrieval.

        If there is already data associated with `session_id`, replace
        it with `data` and reset its expiry time

        Args:
            session_id: The session-ID.
            data: Serialized session data

        Returns:
            None
        """
        async with self._create_sa_session() as sa_session:
            session_obj = await self._get_session_obj(sa_session=sa_session, session_id=session_id)

            if not session_obj:
                session_obj = self._model(session_id=session_id)  # type: ignore[call-arg]
                sa_session.add(session_obj)
            session_obj.data = data
            self._update_session_expiry(session_obj)
            await sa_session.commit()

    async def delete(self, session_id: str) -> None:
        """Delete the data associated with `session_id`. Fails silently if no such session-ID exists.

        Args:
            session_id: The session-ID

        Returns:
            None
        """
        async with self._create_sa_session() as sa_session:
            await sa_session.execute(sa.delete(self._model).where(self._model.session_id == session_id))
            await sa_session.commit()

    async def delete_all(self) -> None:
        """Delete all session data.

        Returns:
            None
        """
        async with self._create_sa_session() as sa_session:
            await sa_session.execute(sa.delete(self._model))
            await sa_session.commit()

    async def delete_expired(self) -> None:
        """Delete all expired session from the database.

        Returns:
            None
        """
        async with self._create_sa_session() as sa_session:
            await sa_session.execute(sa.delete(self._model).where(self._model.expired))
            await sa_session.commit()


class SQLAlchemyBackend(BaseSQLAlchemyBackend[SASession]):
    """Synchronous SQLAlchemy backend."""

    def _get_session_obj(self, *, sa_session: SASession, session_id: str) -> Optional[SessionModelMixin]:
        return sa_session.scalars(self._select_session_obj(session_id)).one_or_none()

    def _get_sync(self, session_id: str) -> Optional[bytes]:
        sa_session = self._create_sa_session()
        session_obj = self._get_session_obj(sa_session=sa_session, session_id=session_id)

        if session_obj:
            if not session_obj.expired:  # type: ignore[truthy-function]
                self._update_session_expiry(session_obj)  # type: ignore[unreachable]
                sa_session.commit()
                return session_obj.data
            sa_session.delete(session_obj)
            sa_session.commit()
        return None

    async def get(self, session_id: str) -> Optional[bytes]:
        """Retrieve data associated with `session_id`.

        Args:
            session_id: The session-ID

        Returns:
            The session data, if existing, otherwise `None`.
        """
        return await run_sync(self._get_sync, session_id)

    def _set_sync(self, session_id: str, data: bytes) -> None:
        sa_session = self._create_sa_session()
        session_obj = self._get_session_obj(sa_session=sa_session, session_id=session_id)

        if not session_obj:
            session_obj = self._model(session_id=session_id)  # type: ignore[call-arg]
            sa_session.add(session_obj)
        session_obj.data = data
        self._update_session_expiry(session_obj)
        sa_session.commit()

    async def set(self, session_id: str, data: bytes) -> None:
        """Store `data` under the `session_id` for later retrieval.

        If there is already data associated with `session_id`, replace
        it with `data` and reset its expiry time

        Args:
            session_id: The session-ID
            data: Serialized session data

        Returns:
            None
        """
        return await run_sync(self._set_sync, session_id, data)

    def _delete_sync(self, session_id: str) -> None:
        sa_session = self._create_sa_session()
        sa_session.execute(sa.delete(self._model).where(self._model.session_id == session_id))
        sa_session.commit()

    async def delete(self, session_id: str) -> None:
        """Delete the data associated with `session_id`. Fails silently if no such session-ID exists.

        Args:
            session_id: The session-ID

        Returns:
            None
        """
        return await run_sync(self._delete_sync, session_id)

    def _delete_all_sync(self) -> None:
        sa_session = self._create_sa_session()

        sa_session.execute(sa.delete(self._model))
        sa_session.commit()

    async def delete_all(self) -> None:
        """Delete all session data.

        Returns:
            None
        """
        await run_sync(self._delete_all_sync)

    def _delete_expired_sync(self) -> None:
        sa_session = self._create_sa_session()
        sa_session.execute(sa.delete(self._model).where(self._model.expired))
        sa_session.commit()

    async def delete_expired(self) -> None:
        """Delete all expired session from the database.

        Returns:
            None
        """
        await run_sync(self._delete_expired_sync)


class SQLAlchemyBackendConfig(ServerSideSessionConfig):
    """Configuration for `SQLAlchemyBackend` and `AsyncSQLAlchemyBackend`"""

    model: Type[SessionModelMixin]
    plugin: SQLAlchemyPlugin

    @validator("plugin", always=True)
    def validate_plugin_config(cls, value: SQLAlchemyPlugin) -> SQLAlchemyPlugin:  # pylint: disable=no-self-argument)
        """Check if the SQLAlchemyPlugin is configured."""
        if not (value._config and value._config.session_maker):
            raise ValueError("Plugin needs to be configured")
        return value

    @property
    def _backend_class(self) -> Type[Union[SQLAlchemyBackend, AsyncSQLAlchemyBackend]]:  # type: ignore[override]
        """Return either `SQLAlchemyBackend` or `AsyncSQLAlchemyBackend`, depending on the engine type configured in the
        `SQLAlchemyPlugin`
        """
        if cast("SQLAlchemyPluginConfig", self.plugin._config).use_async_engine:
            return AsyncSQLAlchemyBackend
        return SQLAlchemyBackend



from starlite.testing.create_test_client import create_test_client
from starlite.testing.request_factory import RequestFactory
from starlite.testing.test_client import TestClient

__all__ = ("TestClient", "RequestFactory", "create_test_client")
